"""
Routines to segment lnPi

 1. find max/peaks in lnPi
 2. segment lnPi about these peaks
 3. determine free energy difference between segments
    a. Merge based on low free energy difference
 4. combination of 1-3.
"""

import warnings
from collections.abc import Iterable
from functools import lru_cache

import bottleneck
import numpy as np
from skimage import feature, morphology, segmentation

from ._docstrings import _prepare_shared_docs, _shared_docs, docfiller
from .lnpienergy import wFreeEnergy
from .lnpiseries import lnPiCollection

# * Common doc strings

_shared_docs_local = {
    "data": """
    data : array-like
        Image data to analyze
    """,
    "min_distance": """
    min_distance : int or sequence of ints, default=(5, 10, 15, 20, 25)
        min_distance parameter.  If sequence, then call
        :func:`~skimage.feature.peak_local_max` until number of peaks ``<=num_peaks_max``.
    """,
    "connectivity_morphology": """
    connectivity : int, optional
        Maximum number of orthogonal hops to consider a pixel/voxel as a neighbor.
        Accepted values are ranging from 1 to ``input.ndim``. If ``None``, a full connectivity of ``input.ndim`` is used.
    """,
    "connectivity_watershed": """
    connectivity : ndarray, optional
        An array with the same number of dimensions as image whose non-zero elements indicate neighbors for connection.
        Following the scipy convention, default is a one-connected array of the dimension of the image.
    """,
    "num_peaks_max": """
    num_peaks_max : int, optional.
        Max number of maxima/peaks to find. If not specified, any number of peaks allowed.
    """,
    "peak_style": """
    style : {'indices', 'mask', 'marker'}
        Controls output style

        * indices : indices of peaks
        * mask : bool array marking peak locations
        * marker : array of ints

    """,
    "markers": """
    markers : int, or ndarray of int, same shape as image, optional
        The desired number of markers, or an array marking the basins with the values to be assigned in the label matrix.
        Zero means not a marker. If None (no markers given), the local minima of the image are used as markers.
    """,
}

_shared_docs_local = _prepare_shared_docs(_shared_docs_local)
docfiller_shared = docfiller(**(dict(_shared_docs, **_shared_docs_local)))

# docfiller_shared = docfiller(**dict(_shared_docs, **_prepare_shared_docs(_shared_docs_local)))


@docfiller_shared
def peak_local_max_adaptive(
    data,
    mask=None,
    min_distance=None,
    threshold_rel=0.00,
    threshold_abs=0.2,
    num_peaks_max=None,
    style="indices",
    connectivity=None,
    errors="warn",
    **kwargs
):
    """
    Find local max with fall backs min_distance and filter.

    This is an adaptation of :func:`~skimage.feature.peak_local_max`, which is
    called iteratively until the number of `peaks` is less than `num_peaks_max`.

    Parameters
    ----------
    {data}
    {mask_image}
    {min_distance}
    threshold_rel, threshold_abs : float
        thresholds parameters
    {num_peaks_max}
    {peak_style}
    {connectivity_morphology}
    indices : bool, optional, default=True
        if True, return indicies of peaks.
        if False, return array of ints of shape `data.shape` with peaks
        marked by value > 0.
    errors : {{'ignore','raise','warn'}}, default='warn'
        - If raise, raise exception if npeaks > num_peaks_max
        - If ignore, return all found maxima
        - If warn, raise warning if npeaks > num_peaks_max
    **kwargs
        Extra arguments to :func:`~skimage.feature.peak_local_max`

    Returns
    -------
    out : indices or mask
        Depending on the value of `indices`.

    Notes
    -----
    The option `mask` is passed as the value `labels` in :func:`~skimage.feature.peak_local_max`

    See Also
    --------
    ~skimage.feature.peak_local_max
    ~skimage.morphology.label
    """

    assert style in ["indices", "mask", "marker"]

    if min_distance is None:
        min_distance = [5, 10, 15, 20, 25]

    if num_peaks_max is None:
        num_peaks_max = np.inf

    if not isinstance(min_distance, Iterable):
        min_distance = [min_distance]

    data = data - bottleneck.nanmin(data)
    kwargs = dict(dict(exclude_border=False), **kwargs)

    n = idx = None
    for md in min_distance:
        idx = feature.peak_local_max(
            data,
            min_distance=md,
            labels=mask,
            threshold_abs=threshold_abs,
            threshold_rel=threshold_rel,
            # this option removed in future
            **kwargs
        )

        n = len(idx)
        if n <= num_peaks_max:
            break

    if n is None or idx is None:
        raise ValueError("failed to find peaks")

    if n > num_peaks_max:
        if errors == "ignore":
            pass
        elif errors in ("raise", "ignore"):
            message = "{} maxima found greater than {}".format(n, num_peaks_max)
            if errors == "raise":
                raise RuntimeError(message)
            else:
                warnings.warn(message)

    idx = tuple(idx.T)
    if style == "indices":
        out = idx
    else:
        out = np.zeros_like(data, dtype=bool)
        out[idx] = True

        if style == "marker":
            out = morphology.label(out, connectivity=connectivity)

    return out


@docfiller_shared
class Segmenter(object):
    """
    Data segmenter:


    Parameters
    ----------
    {min_distance}
    {peak_kws}
    {watershed_kws}
    """

    def __init__(self, min_distance=None, peak_kws=None, watershed_kws=None):

        if min_distance is None:
            min_distance = [1, 5, 10, 15, 20]

        if peak_kws is None:
            peak_kws = {}
        self.peak_kws = peak_kws

        if watershed_kws is None:
            watershed_kws = {}
        self.watershed_kws = watershed_kws

    @docfiller_shared
    def peaks(self, data, mask=None, num_peaks_max=None, style="marker", **kwargs):
        """
        Interface to :func:`peak_local_max_adaptive` with default values from `self`.


        Parameters
        ----------
        {data}
        {num_peaks_max}
        {peak_style}
        {mask_image}

        **kwargs
            Extra arguments to :func:`peak_local_max_adaptive`.

        Returns
        -------
        out :
            If ``style=='marker'``, then return label array.  Otherwise,
            return indicies of peaks.

        Notes
        -----
        Any value not set will be inherited from `self.peak_kws`


        See Also
        --------
        peak_local_max_adaptive

        """

        if mask is not None:
            kwargs["mask"] = mask
        if num_peaks_max is not None:
            kwargs["num_peaks_max"] = num_peaks_max
        kwargs["style"] = style
        kwargs = dict(self.peak_kws, **kwargs)
        out = peak_local_max_adaptive(data, **kwargs)

        return out

    @docfiller_shared
    def watershed(self, data, markers, mask, connectivity=None, **kwargs):
        """
        Interface to :func:`skimage.segmentation.watershed` function

        Parameters
        ----------
        {data}
        {markers}

        {mask_image}
        connectivity : int
            connectivity to use in watershed
        {connectivity_watershed}
        **kwargs
            Extra arguments to :func:`~skimage.segmentation.watershed`

        Returns
        -------
        labels : array of ints
            Values > 0 correspond to found regions

        See Also
        --------
        ~skimage.segmentation.watershed
        """

        if connectivity is None:
            connectivity = data.ndim
        kwargs = dict(self.watershed_kws, connectivity=connectivity, *kwargs)
        return segmentation.watershed(data, markers=markers, mask=mask, **kwargs)

    @docfiller_shared
    def segment_lnpi(
        self,
        lnpi,
        markers=None,
        find_peaks=True,
        num_peaks_max=None,
        connectivity=None,
        peaks_kws=None,
        watershed_kws=None,
    ):
        """
        Perform segmentations of lnPi object using watershed on negative of lnPi data.

        Parameters
        ----------
        lnPi : lnPiMasked
            Object to be segmented
        {markers}

        find_peaks : bool, default=True
            If True, use :func:`peak_local_max_adaptive` to construct `markers`.
        {num_peaks_max}
        {connectivity_watershed}
        {peak_kws}
        {watershed_kws}

        Returns
        -------
        {labels}


        See Also
        --------
        Segmenter.watershed
        skimage.morphology.watershed

        """

        if markers is None:
            if find_peaks:
                if peaks_kws is None:
                    peaks_kws = {}
                markers = self.peaks(
                    lnpi.data,
                    mask=~lnpi.mask,
                    num_peaks_max=num_peaks_max,
                    connectivity=connectivity,
                    **peaks_kws
                )
            else:
                markers = num_peaks_max

        if watershed_kws is None:
            watershed_kws = {}
        labels = self.watershed(
            -lnpi.data, markers=markers, mask=~lnpi.mask, connectivity=connectivity
        )
        return labels


class PhaseCreator(object):
    """
    Helper class to create phases

    Parameters
    ----------
    nmax : int
        Maximum number of phases to allow
    nmax_peak : int, optional
        if specified, the allowable number of peaks to locate.
        This can be useful for some cases.  These phases will be merged out at the end.
    ref : lnPiMasked, optional
    segmenter : Segmenter object, optional
        segmenter object to create labels/masks. Defaults to using base segmenter.
    segment_kws : mapping, optional
        Optional arguments to be passed to :meth`Segmenter.segmenter_lnpi`.
    tag_phases : callable, optional
        Optional funciton which takes a list of :class:`lnpy.lnPiMasked` objects
        and returns on integer label for each object.
    phases_factory : callable, optional
        Factory function for returning Collection from a list of :class:`lnpy.lnPiMasked` object.
        Defaults to :meth:`lnpy.lnPiCollection.from_list`.
    lnPiFreeEnergy_kws : mapping, optional
        Optional arguments to ...
    merge_kws : mapping, optional
        Optional arguments to ...

    """

    def __init__(
        self,
        nmax,
        nmax_peak=None,
        ref=None,
        segmenter=None,
        segment_kws=None,
        tag_phases=None,
        phases_factory=None,
        lnPiFreeEnergy_kws=None,
        merge_kws=None,
    ):

        if phases_factory is None:
            phases_factory = lnPiCollection.from_list

        if nmax_peak is None:
            nmax_peak = nmax * 2
        self.nmax = nmax
        self.ref = ref

        if segmenter is None:
            segmenter = Segmenter()
        self.segmenter = segmenter

        self.tag_phases = tag_phases

        self.phases_factory = phases_factory

        if segment_kws is None:
            segment_kws = {}
        self.segment_kws = segment_kws
        self.segment_kws["num_peaks_max"] = nmax_peak

        if lnPiFreeEnergy_kws is None:
            lnPiFreeEnergy_kws = {}
        self.lnPiFreeEnergy_kws = lnPiFreeEnergy_kws

        if merge_kws is None:
            merge_kws = {}
        merge_kws = dict(merge_kws, convention=False, nfeature_max=self.nmax)
        self.merge_kws = merge_kws

    def _merge_phase_ids(sel, ref, phase_ids, lnpis):
        """
        perform merge of phase_ids/index
        """
        from scipy.spatial.distance import pdist

        if len(phase_ids) == 1:
            # only single phase_id
            return phase_ids, lnpis

        phase_ids = np.array(phase_ids)
        dist = pdist(phase_ids.reshape(-1, 1)).astype(int)
        if not np.any(dist == 0):
            # all different
            return phase_ids, lnpis

        phase_ids_new = []
        masks_new = []
        for idx in np.unique(phase_ids):
            where = np.where(idx == phase_ids)[0]
            mask = np.all([lnpis[i].mask for i in where], axis=0)

            phase_ids_new.append(idx)
            masks_new.append(mask)
        lnpis_new = ref.list_from_masks(masks_new, convention=False)

        return phase_ids_new, lnpis_new

    def build_phases(
        self,
        lnz=None,
        ref=None,
        efac=None,
        nmax=None,
        nmax_peak=None,
        connectivity=None,
        reweight_kws=None,
        merge_phase_ids=True,
        merge_phases=True,
        phases_factory=None,
        phase_kws=None,
        segment_kws=None,
        lnPiFreeEnergy_kws=None,
        merge_kws=None,
        tag_phases=None,
    ):
        """
        Construct 'phases' for a lnPi object.

        This is quite an involved process.  The steps are

        * Optionally find the location of the maxima in lnPi.
        * Segment lnpi using watershed
        * Merge phases which are energentically similar
        * Optionally merge phases which have the same `phase_id`

        Parameters
        ----------
        lnz : int or sequence of ints, optional
            lnz value to evaluate `ref` at.  If not specified, use
            `ref.lnz`
        ref : lnPiMasked object
            Object to be segmented
        efac : float, optional
            Optional value to use in energetic merging of phases.
        nmax : int, optional
            Maximum number of phases.  Defaults to `self.nmax`
        nmax_peak : int, optional
            Maximum number of peaks to allow in :func:`peak_local_max_adaptive`.
            Note that this value can be larger than `nmax`.  Defaults to `self.nmax_peak`.
        {connectivity_morphology}
        reweight_kws : mapping, optional
            Extra arguments to `ref.reweight`
        merge_phase_ids : bool, default = True
            If True and calling `tag_phases` routine, merge phases with same phase_id.
        phases_factory : callable, optional
            Function to convert list of phases into Phases object.
            Defaults to `self.phases_factory`.
        phase_kws : mapping, optional
            Extra arguments to `phases_factory`
        segment_kws : mapping, optional
            Extra arguments to `self.segmenter.segment_lnpi`
        lnPiFreeEnergy_kws : mapping, optional
            Extra arguments to free energy calculation
        merge_kws : mapping, optional
            Extra arguments to merge
        tag_phases : callable, optional
            Funciton to tag phases.  Defaults to `self.tag_phases`

        """

        def _combine_kws(class_kws, passed_kws, **default_kws):
            if class_kws is None:
                class_kws = {}
            if passed_kws is None:
                passed_kws = {}
            return dict(class_kws, **default_kws, **passed_kws)

        if ref is None:
            if self.ref is None:
                raise ValueError("must specify ref or self.ref")
            ref = self.ref

        # reweight
        if lnz is not None:
            if reweight_kws is None:
                reweight_kws = {}
            ref = ref.reweight(lnz, **reweight_kws)

        if nmax is None:
            nmax = self.nmax

        if nmax_peak is None:
            nmax_peak = nmax * 2

        connectivity_kws = {}
        if connectivity is not None:
            connectivity_kws["connectivity"] = connectivity

        if nmax > 1:
            # segment lnpi using watershed
            segment_kws = _combine_kws(
                self.segment_kws,
                segment_kws,
                num_peaks_max=nmax_peak,
                **connectivity_kws
            )
            labels = self.segmenter.segment_lnpi(lnpi=ref, **segment_kws)

            # analyze w = - lnPi
            lnPiFreeEnergy_kws = _combine_kws(
                self.lnPiFreeEnergy_kws, lnPiFreeEnergy_kws, **connectivity_kws
            )
            wlnpi = wFreeEnergy.from_labels(
                data=ref.data, labels=labels, **lnPiFreeEnergy_kws
            )

            if merge_phases:
                # merge
                other_kws = {} if efac is None else {"efac": efac}
                merge_kws = _combine_kws(
                    self.merge_kws, merge_kws, nfeature_max=nmax, **other_kws
                )
                masks, _, _ = wlnpi.merge_regions(**merge_kws)
            else:
                masks = wlnpi.masks

            # list of lnpi
            lnpis = ref.list_from_masks(masks, convention=False)

            # tag phases?
            if tag_phases is None:
                tag_phases = self.tag_phases
            if tag_phases is not None:
                index = tag_phases(lnpis)
                if merge_phase_ids:
                    index, lnpis = self._merge_phase_ids(ref, index, lnpis)
            else:
                index = list(range(len(lnpis)))
        else:
            lnpis = [ref]
            index = [0]

        if phases_factory is None:
            phases_factory = self.phases_factory
        if isinstance(phases_factory, str) and phases_factory.lower() == "none":
            phases_factory = None
        if phases_factory is not None:
            if phase_kws is None:
                phase_kws = {}
            return phases_factory(items=lnpis, index=index, **phase_kws)
        else:
            return lnpis, index

    def build_phases_mu(self, lnz):
        """
        Create constructor at fixed values of `mu`

        See Also
        --------
        BuildPhases_mu
        """
        return BuildPhases_mu(lnz, self)

    def build_phases_dmu(self, dlnz):
        """
        Create constructor at fixed values of `dmu`.

        See Also
        --------
        BuildPhases_dmu

        """
        return BuildPhases_dmu(dlnz, self)


class BuildPhasesBase(object):
    """
    Base class to build Phases objecs from scalar values of `lnz`.
    """

    def __init__(self, X, phase_creator):
        self._phase_creator = phase_creator
        self.X = X

    @property
    def X(self):
        return self._X

    @property
    def phase_creator(self):
        return self._phase_creator

    @X.setter
    def X(self, X):
        assert sum([x is None for x in X]) == 1
        self._X = X
        self._ncomp = len(self._X)
        self._index = self._X.index(None)
        self._set_params()

    @property
    def index(self):
        return self._index

    def _set_params(self):
        pass

    def _get_lnz(self, lnz_index):
        # to be implemented in child class
        raise NotImplementedError

    def __call__(self, lnz_index, *args, **kwargs):
        """
        Build phases from scalar value of lnz.
        """
        lnz = self._get_lnz(lnz_index)
        return self._phase_creator.build_phases(lnz=lnz, *args, **kwargs)


# from .utils import get_lnz_iter
class BuildPhases_mu(BuildPhasesBase):
    """
    create phases from scalar value of mu for fixed value of mu for other species

    Parameters
    ----------
    lnz : list
        list with one element equal to None.  This is the component which will be varied
        For example, lnz=[lnz0, None, lnz2] implies use values of lnz0,lnz2 for components 0 and 2, and
        vary component 1
    phase_creator : PhaseCreator object
    """

    def __init__(self, lnz, phase_creator):
        super().__init__(X=lnz, phase_creator=phase_creator)

    def _get_lnz(self, lnz_index):
        lnz = self.X.copy()
        lnz[self.index] = lnz_index
        return lnz


class BuildPhases_dmu(BuildPhasesBase):
    """
    Create phases from scalar value of mu at fixed value of dmu for other species

    Parameters
    ----------
    dlnz : list
        list with one element equal to None.  This is the component which will be varied
        For example, dlnz=[dlnz0,None,dlnz2] implies use values of dlnz0,dlnz2
        for components 0 and 2, and vary component 1.
        dlnz_i = lnz_i - lnz_index, where lnz_index is the value varied.
    phase_creator : :class:`lnpy.PhaseCreator`
    """

    def __init__(self, dlnz, phase_creator):
        super().__init__(X=dlnz, phase_creator=phase_creator)

    def _set_params(self):
        self._dlnz = np.array([x if x is not None else 0.0 for x in self.X])

    def _get_lnz(self, lnz_index):
        return self._dlnz + lnz_index


@lru_cache(maxsize=10)
def get_default_PhaseCreator(nmax):
    return PhaseCreator(nmax=nmax)



from __future__ import absolute_import, division, print_function

from functools import lru_cache, wraps

import numpy as np

# import pandas as pd
import xarray as xr

from .cached_decorators import gcached_use_cache as gcached
from .lnpidata import MaskedlnPiDelayed, lnPiMasked
from .lnpiseries import lnPiCollection
from .maskedlnpi_legacy import MaskedlnPiLegacy
from .utils import dim_to_suffix_dataset


###############################################################################
# xlnPi wrapper
@lru_cache(maxsize=10)
def _get_indices(shape):
    return np.indices(shape)


@lru_cache(maxsize=10)
def _get_range(n):
    return np.arange(n)


@lru_cache(maxsize=10)
def get_xrlnPiWrapper(
    shape,
    rec_name="sample",
    n_name="n",
    lnz_name="lnz",
    comp_name="component",
    phase_name="phase",
):
    return xlnPiWrapper(
        shape=shape,
        rec_name=rec_name,
        n_name=n_name,
        lnz_name=lnz_name,
        comp_name=comp_name,
        phase_name=phase_name,
    )


class xlnPiWrapper:
    _use_cache = True
    """
    This class just wraps lnPi with xarray functionality

    Most likely, this shouldn't be accessed by the user.


    """

    def __init__(
        self,
        shape,
        rec_name="sample",
        n_name="n",
        lnz_name="lnz",
        comp_name="component",
        phase_name="phase",
    ):

        self.shape = shape
        self.ndim = len(shape)

        # dimension names
        if rec_name is None:
            self.dims_rec = []
        else:
            self.dims_rec = [rec_name]

        self.n_name = n_name
        self.dims_n = ["{}_{}".format(n_name, i) for i in range(self.ndim)]
        self.dims_lnz = ["{}_{}".format(lnz_name, i) for i in range(self.ndim)]
        self.dims_comp = [comp_name]

    @gcached()
    def coords_n(self):
        return {
            k: xr.DataArray(
                _get_range(n), dims=k, attrs={"long_name": r"${}$".format(k)}
            )
            for k, n in zip(self.dims_n, self.shape)
        }

    @gcached(prop=False)
    def attrs(self, *args):
        d = {
            "dims_n": self.dims_n,
            "dims_lnz": self.dims_lnz,
            "dims_comp": self.dims_comp,
            "dims_state": self.dims_lnz + list(args),
        }

        if self.dims_rec is not None:
            d["dims_rec"] = self.dims_rec
        return d

    @gcached(prop=False)
    def ncoords(self, coords_n=False):
        if coords_n:
            coords = self.coords_n
        else:
            coords = None
        return xr.DataArray(
            _get_indices(self.shape),
            dims=self.dims_comp + self.dims_n,
            coords=coords,
            name=self.n_name,
        )

    @gcached(prop=False)
    def ncoords_tot(self, coords_n=False):
        return self.ncoords(coords_n).sum(self.dims_comp)

    def wrap_data(self, data, coords_n=False, coords=None, **kwargs):
        if coords is None:
            coords = {}

        if coords_n:
            coords = dict(coords, **self.coords_n)

        return xr.DataArray(
            data,
            dims=self.dims_rec + self.dims_n,
            coords=coords,
            attrs=self.attrs(*kwargs.keys()),
        )

    def wrap_lnpi(self, lnpi, coords_n=False, **kwargs):
        return self.wrap_data(lnpi, coords_n=coords_n, **kwargs)

    def wrap_lnz(self, lnz, **kwargs):
        return xr.DataArray(lnz, dims=self.dims_rec + self.dims_comp, **kwargs)

    def wrap_lnpi_0(self, lnpi_0, **kwargs):
        return xr.DataArray(lnpi_0, dims=self.dims_rec, **kwargs)


################################################################################
# xlnPi
# register xge property to collection classes
def xr_name(long_name=None, name=None, unstack=True, **kws):
    """
    decorator to add name, longname to xarray output
    """

    def decorator(func):
        if name is None:
            _name = func.__name__.lstrip("_")

        else:
            _name = name

        @wraps(func)
        def wrapper(self, *args, **kwargs):
            out = func(self, *args, **kwargs).rename(_name)
            if hasattr(self, "_standard_attrs"):
                attrs = self._standard_attrs
            else:
                attrs = {}
            attrs = dict(attrs, **kws)

            if long_name is not None:
                attrs["long_name"] = long_name
            out = out.assign_attrs(**attrs)
            if unstack and self._xarray_unstack:
                out = out.unstack()

            return out

        return wrapper

    return decorator


@lnPiMasked.decorate_accessor("xge")
@MaskedlnPiDelayed.decorate_accessor("xge")
@MaskedlnPiLegacy.decorate_accessor("xge")
@lnPiCollection.decorate_accessor("xge")
def xge_accessor(parent):
    """
    Accessor to :class:`~lnpy.xGrandCanonical`.
    """
    return xGrandCanonical(parent)


class xGrandCanonical:
    """
    :class:`DataArray` accessor to Grand Canonical properties from lnPi


    This class is primarily interacted with through the attributes ``xge`` attached
    to :class:`~lnpy.lnPiMasked` and :class:`~lnpy.lnPiCollection`.
    """

    def __init__(self, parent):
        self._parent = parent

        self._rec_name = getattr(self._parent, "_concat_dim", None)
        if self._rec_name is None:
            shape = self._parent.shape
        else:
            shape = self._parent._series.iloc[0].shape

        self._wrapper = get_xrlnPiWrapper(shape=shape, rec_name=self._rec_name)

    @property
    def _use_cache(self):
        return getattr(self._parent, "_use_cache", False)

    @property
    def _xarray_unstack(self):
        return getattr(self._parent, "_xarray_unstack", True)

    @gcached()
    def _standard_attrs(self):
        return self._wrapper.attrs(*self._parent.state_kws.keys())

    # @gcached() to much memory
    @property
    def _rec_coords(self):
        if self._rec_name is None:
            return dict(self._parent._index_dict(), **self._parent.state_kws)
        else:
            return {
                self._parent._concat_dim: self._parent.index,
                **self._parent.state_kws,
            }

    @property
    def _xarray_dot_kws(self):
        return getattr(self._parent, "_xarray_dot_kws", {})

    @property
    def ncoords(self):
        """Particle number coordinates."""
        return self._wrapper.ncoords()

    @property
    def ncoords_tot(self):
        """Total number of particles coordinates"""
        return self._wrapper.ncoords_tot()

    @property
    def dims_n(self):
        """Dimension(s) corresponding to  number of particles"""
        return self._wrapper.dims_n

    @property
    def dims_lnz(self):
        r"""Dimension corresponding to values of :math:`\ln z`"""
        return self._wrapper.dims_lnz

    @property
    def dims_comp(self):
        """Dimension of component number"""
        return self._wrapper.dims_comp

    @property
    def dims_state(self):
        """Dimensions corresponding to 'state variables'"""
        return self.pi_norm.attrs["dims_state"]

    @property
    def dims_rec(self):
        """Dimensions for replicates"""
        return self._wrapper.dims_rec

    @property
    def beta(self):
        r"""Inverse temperature :math:`\beta = 1 / (k_{\rm B} T)`"""
        return self._parent.state_kws["beta"]

    @property
    def volume(self):
        """System volume :math:`V`."""
        return self._parent.state_kws["volume"]

    @gcached()
    def coords_state(self):
        return {k: self.pi_norm.coords[k] for k in self.dims_state}

    @gcached()
    @xr_name(r"$\beta {\bf \mu}$")
    def betamu(self):
        r"""Scaled checmical potential :math:`\beta \mu`"""
        return self._wrapper.wrap_lnz(self._parent._lnz_tot, coords=self._rec_coords)

    @property
    @xr_name(r"$\ln\beta{\bf\mu}$")
    def lnz(self):
        r"""Log of activity :math:`\ln z``"""
        return self.betamu

    @xr_name(r"$\ln \Pi(n,\mu,V,T)$", unstack=False)
    def lnpi(self, fill_value=None):
        r"""DataArray view of :math:`\ln \Pi(N)`

        Notes
        -----
        This value is always in 'stacked' form.
        You must manually unstack it.
        """
        return self._wrapper.wrap_lnpi(
            self._parent._lnpi_tot(fill_value),
            coords=self._rec_coords,
            **self._parent.state_kws,
        )  # .assign_coords(**self._rec_coords)

    @gcached()
    def _pi_params(self):
        pi_norm, pi_sum, lnpi_zero = self._parent._pi_params(-np.inf)

        pi_sum = self._wrapper.wrap_lnpi_0(
            pi_sum, name="pi_sum", coords=self._rec_coords
        )

        lnpi_zero = self._wrapper.wrap_lnpi_0(
            lnpi_zero, name="lnpi_zero", coords=self._rec_coords
        )

        pi_norm = self._wrapper.wrap_lnpi(
            pi_norm, coords=self._rec_coords, **self._parent.state_kws
        )

        return pi_norm, pi_sum, lnpi_zero

    @property
    def pi_norm(self):
        r"""Normalized value of :math:`\Pi_{\rm norm}(N) = \Pi(N) / \sum_N \Pi(N)`"""
        return self._pi_params[0]

    @property
    def pi_sum(self):
        r"""Sum of unnormalized :math:`\Pi(N)`"""
        return self._pi_params[1]

    @property
    def _lnpi_zero(self):
        return self._pi_params[2]

    @property
    def lnpi_norm(self):
        r""":math:`\ln \Pi_{\rm norm}(N)`."""
        pi = self.pi_norm
        return np.log(xr.where(pi > 0, pi, np.nan))

    def _get_prop_from_extra_kws(self, prop):
        if isinstance(prop, str):
            p = self._parent
            if hasattr(self._parent, "_series"):
                p = p.iloc[0]

            if prop not in p.extra_kws:
                raise ValueError(f"{prop} not found in extra_kws")
            else:
                prop = p.extra_kws.get(prop, None)

        return prop

    def _array_or_callable_to_xarray(self, x, allow_extra_kws=True, *args, **kwargs):

        if callable(x):
            x = x(self, *args, **kwargs)
        elif allow_extra_kws:
            x = self._get_prop_from_extra_kws(x)

        if not isinstance(x, xr.DataArray):
            x = xr.DataArray(x, dims=self.dims_n)

        return x

    def _mean_pi(self, x, *args, **kwargs):

        return xr.dot(self.pi_norm, x, dims=self.dims_n, **self._xarray_dot_kws)

    @xr_name()
    def mean_pi(self, x, *args, **kwargs):
        r"""
        Calculates :math:`\overline{x} = \sum_N \Pi_{\rm norm}(N) x(N)`

        Parameters
        ----------
        x : array, callable, or str
            If callable, should have form ``x = x(self, *args, **kwargs)``.
            If string, then set `x = self.parent.extra_kws[x]`.
            Otherwise, should be an array of same shape as single lnPi.
        *args, **kwargs
            Extra arguments to `x` if passing callable
        Returns
        -------
        ave : DataArray
        x can be an array or a callable

        f(self, *args, **kwargs)

        if x is not an xr.DataArray, try to convert it to one.
        """
        x = self._array_or_callable_to_xarray(x, *args, **kwargs)
        return self._mean_pi(x, *args, **kwargs)

    def _central_moment_bar(
        self,
        x,
        y=None,
        xmom=2,
        ymom=None,
        xmom_dim="xmom",
        ymom_dim="ymom",
        *args,
        **kwargs,
    ):
        """
        calculate central moments of the form sum(Pi * (\bar{x} - <\bar{x}>) ** n)

        Note that if you pass in the

        """

        def _get_dx(xx, mom, mom_dim):
            xx = self._array_or_callable_to_xarray(xx, *args, **kwargs)
            if isinstance(mom, int):
                mom = [mom]

            mom = xr.DataArray(mom, dims=mom_dim, coords={mom_dim: mom})

            return (xx - self._mean_pi(xx)) ** mom

        dx = _get_dx(x, xmom, xmom_dim)

        if y is not None:
            dy = _get_dx(y, ymom, ymom_dim)

            dx = dx * dy

        return self._mean_pi(dx)

    def var_pi(self, x, y=None, *args, **kwargs):
        r"""Calculate Grand Canonical variance from canonical properties.

        Given x(N) and y(N), calculate

        .. math::

            {\rm var}(x, y) = \overline{(x - \overline{x}) (y - \overline{y})}

        ``x`` and ``y`` can be arrays, or callables, in which case:
        ``x = x(self, *args, **kwargs)``

        See Also
        --------
        :math:`mean_pi`

        """

        if callable(x):
            x = x(self, *args, **kwargs)

        # x_mean = self.mean_pi(x)
        # if y is None:
        #     y = x
        # if y is x:
        #     y_mean = x_mean
        # else:
        #     if callable(y):
        #         y = y(self, *args, **kwargs)
        #     y_mean = self.mean_pi(y)
        # return self.mean_pi((x - x_mean) * (y - y_mean))

        # this cuts down on memory usage
        xx = x - self._mean_pi(x)
        if y is None:
            yy = xx
        else:
            if callable(y):
                y = y(self, *args, **kwargs)
            yy = y - self._mean_pi(y)

        return xr.dot(self.pi_norm, xx, yy, dims=self.dims_n, **self._xarray_dot_kws)

    def pipe(self, func, *args, **kwargs):
        """
        Apply function to `self`
        """
        return func(self, *args, **kwargs)

    @gcached()
    @xr_name(r"${\bf n}(\mu,V,T)$")
    def nvec(self):
        r"""Average number of particles of each component :math:`\overline{{\bf N}}`"""
        return self._mean_pi(self.ncoords)

    @gcached()
    @xr_name(r"$n(\mu,V,T)$")
    def ntot(self):
        r"""Average total number of particles :math:`\overline{N}`"""
        return self._mean_pi(self.ncoords_tot)

    @property
    @xr_name(r"${\bf x}(\mu,V,T)$")
    def molfrac(self):
        r"""Average molfrac for each components :math:`{\bf x} = \overline{{\bf N}} / N`"""
        return self.nvec.pipe(lambda x: x / x.sum(self.dims_comp))

    @gcached()
    @xr_name(r"$var[{\bf n}(\mu,V,T)]$")
    def nvec_var(self):
        r"""Variance in particle number :math:`{\rm var}\, \bf{N}`"""
        return self.var_pi(self.ncoords)

    @gcached()
    @xr_name(r"$var[n(\mu,V,T)]$")
    def ntot_var(self):
        r"""Variance in total number of particles :math:`{\rm var}\, N`"""
        return self.var_pi(self.ncoords_tot)

    @property
    @xr_name(r"${\bf \rho}(\mu,V,T)$")
    def dens(self):
        r"""Density of each component :math:`{\bf \rho} = \overline{{\bf N}} / V`"""
        # NOTE: keep this here because of some internal calculations
        return self.nvec.pipe(lambda x: x / x["volume"])

    @property
    @xr_name(r"$\rho(\mu,V,T)$", standard_name="total_density")
    def dens_tot(self):
        r"""Total density :math:`\overline{N} / V`"""
        return self.ntot.pipe(lambda x: x / x["volume"])

    def max(self, *args, **kwargs):
        r""":math:`\max_N \ln \Pi(N)`"""
        return self.lnpi().max(self.dims_n)

    # def argmax(self, *args, **kwargs):
    #     return np.array(
    #         [x.local_argmax(*args, **kwargs) for x in self._parent])

    # @xr_name('distance from upper edge')
    # def edge_distance(self, ref, *args, **kwargs):
    #     """distance from endpoint"""

    #     return xr.DataArray([x.edge_distance(ref) for x in self._parent],
    #                         dims=self.dims_rec,
    #                         coords=self._rec_coords)

    @gcached(prop=False)
    def _argmax_indexer(self):
        x = self.pi_norm.values
        xx = x.reshape(x.shape[0], -1)
        idx_flat = xx.argmax(-1)
        indexer = np.unravel_index(idx_flat, x.shape[1:])
        return indexer

    @gcached()
    def _argmax_indexer_dict(self):
        return {
            k: xr.DataArray(v, dims=self._parent._concat_dim)
            for k, v in zip(self.dims_n, self._argmax_indexer())
        }

    @gcached()
    def _sample_indexer_dict(self):
        return {
            self._parent._concat_dim: xr.DataArray(
                range(len(self._parent)), dims=self._parent._concat_dim
            )
        }

    @property
    def _sample_argmax_indexer_dict(self):
        return dict(self._argmax_indexer_dict, **self._sample_indexer_dict)

    def lnpi_max(self, fill_value=None, add_n_coords=True):
        r"""maximum value of :math:`\max_N \ln \Pi(N, ...)`"""
        out = self.lnpi(fill_value).isel(**self._sample_argmax_indexer_dict)

        # NOTE : This assumes each n value corresponds to index
        # alternatively, could put through filter like
        # coords = {k : out[k].isel(**{k : v}) for k, v in self._argmax_index_dict.items()}

        if add_n_coords:
            out = out.assign_coords(**self._argmax_indexer_dict)

        return out

    def pi_norm_max(self, add_n_coords=True):
        r"""Maximum value :math:`\max_N \Pi_{\rm norm}(N, meta)`"""
        out = self.pi_norm.isel(**self._sample_argmax_indexer_dict)
        if add_n_coords:
            out = out.assign_coords(**self._argmax_indexer_dict)
        return out

    @gcached(prop=False)
    def argmax(self):
        return np.array(self._argmax_indexer()).T

    @xr_name("distance from upper edge")
    def edge_distance(self, ref, *args, **kwargs):
        """Distance from argmax(lnPi) to endpoint"""
        out = ref.edge_distance_matrix[self._argmax_indexer()]
        return xr.DataArray(out, dims=self.dims_rec, coords=self._rec_coords)

    @xr_name("distance from edge of cut value")
    def edge_distance_val(self, ref, val=None, max_frac=None, *args, **kwargs):
        """
        Calulate min distance from where self.pi_norm > val to edge

        Parameters
        ----------
        ref : lnPiMasked
            reference object to consider.
        val : float

        max_frac : bool, optional
        if not None, val = max_frac * self.pi_norm.max(self.dims_n)
        """

        if max_frac is not None:
            assert max_frac < 1.0
            val = self.pi_norm_max(add_n_coords=False) * max_frac
        else:
            assert val is not None

        e = xr.DataArray(ref.edge_distance_matrix, dims=self.dims_n)
        mask = self.pi_norm > val
        return e.where(mask).min(self.dims_n)

    @gcached(prop=False)
    @xr_name(r"$\beta \Omega(\mu,V,T)$", standard_name="grand_potential")
    def _betaOmega(self, lnpi_zero=None):
        if lnpi_zero is None:
            lnpi_zero = self._lnpi_zero
        return lnpi_zero - np.log(self.pi_sum)

    def betaOmega(self, lnpi_zero=None):
        r"""Scaled grand potential :math:`\beta \Omega`.

        Parameters
        ----------
        lnpi_zero : float or None
            if None, lnpi_zero = self.data.ravel()[0]
        """
        # Note.  Put calculation in _betaOmega
        # because so many other things
        # call it with lnpi_zero set to None
        # so end up with copies of the same
        # thing in cache
        return self._betaOmega(lnpi_zero)

    @gcached()
    @xr_name(r"${\rm PE}(\mu,V,T)$", standard_name="potential_energy")
    def PE(self):
        r"""Potential energy :math:`\overline{PE}`"""

        # if betaPE available, use that:

        if hasattr(self._parent, "_series"):
            PE = self._parent.iloc[0].extra_kws.get("PE", None)
        else:
            PE = self._parent.extra_kws.get("PE", None)

        if PE is None:
            raise AttributeError('must set "PE" in "extra_kws" of lnPiMasked')
        else:
            PE = xr.DataArray(PE, dims=self.dims_n)
        return self._mean_pi(PE)

    @xr_name(r"$\beta F(\mu,V,T)$", standard_name="helmholtz_free_energy")
    def betaF_alt(self, betaF_can, correction=True):
        r"""Alternate scaled Helmholtz free energy :math:`\beta \overline{F}`.

        Calculated using

        .. math::

            \beta \overline{F} = \sum_N [\Pi(N) \beta F(N) + C(N)]

        Parameters
        ----------
        betaF_can : array-like
            Value of :math:`F(N)`
        correction : bool, default=True
            If True, :math:`C(N) = \ln \Pi(N) `.  Otherwise, :math:`C=0`.
        """
        if correction:
            betaF_can = betaF_can + self.lnpi_norm
        # return (self.pi_norm * betaF_can).sum(self.dims_n)
        return self._mean_pi(betaF_can)

    ################################################################################
    # Other properties
    @xr_name(r"$\beta\omega(\mu,V,T)$", standard_name="grand_potential_per_particle")
    def betaOmega_n(self, lnpi_zero=None):
        r"""Grand potential per particle :math:`\beta \Omega / \overline{N}`"""
        return self.betaOmega(lnpi_zero) / self.ntot

    # @gcached(prop=False)
    @xr_name(r"$\beta p(\mu,V,T)V$")
    def betapV(self, lnpi_zero=None):
        r""":math:`\beta p V = - \beta \Omega`"""
        return -self.betaOmega(lnpi_zero)

    @gcached()
    @xr_name("mask_stable", description="True where state is most stable")
    def mask_stable(self):
        """
        Masks are True where values are stable. Only works for unstacked data.
        """
        if not self._xarray_unstack:
            # raise Value'only mask with unstack')
            pv = self.betapV()
            sample = self._parent._concat_dim
            return (
                pv.unstack(sample)
                .pipe(lambda x: x.max("phase") == x)
                .stack(sample=pv.indexes[sample].names)
                .loc[pv.indexes["sample"]]
            )
        else:
            return self.betapV().pipe(lambda x: x.max("phase") == x)

    # @gcached(prop=False)
    @xr_name(r"$\beta p(\mu,V,T)/\rho$", standard_name="compressibility_factor")
    def Z(self, lnpi_zero=None):
        r"""Compressibility factor :math:`\beta p / \rho`"""
        return -self.betaOmega_n(lnpi_zero)

    @xr_name(r"$p(\mu,V,T)$")
    def pressure(self, lnpi_zero=None):
        r"""Pressure :math:`p = -\Omega / V`"""
        return self.betapV(lnpi_zero).pipe(lambda x: x / (x["beta"] * x["volume"]))

    @property
    @xr_name(r"${\rm PE}(\mu,V,T)/n$", standard_name="potential_energy_per_particle")
    def PE_n(self):
        r"""Potential energy per particle :math:`\overline{PE}/\overline{N}`"""
        return self.PE / self.ntot

    def table(
        self,
        keys=None,
        default_keys=("nvec", "betapV", "PE_n"),
        ref=None,
        mask_stable=False,
        dim_to_suffix=None,
    ):
        """
        Create :class:`xarray.Dataset` of calculated properties.

        Parameters
        ----------
        keys : sequence of str, optional
            keys of attributes or methods of `self` to include in ouput
        default_keys : sequence of str
            Default keys to consider.
        ref : lnPiMasked, optional
            If calculating `edge_distastance`, need a reference :class:`~lnpy.lnPiMasked` object.
        mask_stable : bool, default=False
            If True, remove any unstable values
        dim_to_suffix : sequence of hashables, optional
            dimensions to remove from ouput.  These are instead added as suffix to variable names

        Returns
        -------
        ds : Dataset
            Contianing all the calculated properties in a single object

        Notes
        -----
        The results can be easily convert to a :class:`pandas.DataFrame` using ``ds.to_frame()``
        """

        out = []
        if ref is not None:
            out.append(self.edge_distance(ref))

        def _process_keys(x):
            if x is None:
                return []
            elif isinstance(x, str):
                return [x]
            else:
                return list(x)

        keys = _process_keys(keys) + _process_keys(default_keys)
        if not keys:
            raise ValueError("must specify some keys or default_keys to use")

        # only unique keys
        # this preserves order
        for key in dict.fromkeys(keys):
            try:
                v = getattr(self, key)
                if callable(v):
                    v = v()
                out.append(v)
            except Exception:
                pass

        out = xr.merge(out)
        if "lnz" in keys:
            # if including property lnz, then drop lnz_0, lnz_1,...
            out = out.drop(out["lnz"]["dims_lnz"])

        if mask_stable:
            # mask_stable inserts nan in non-stable
            mask_stable = self.mask_stable
            if not self._xarray_unstack:
                out = out.where(mask_stable, drop=True)
            else:
                phase = out.phase
                out = (
                    out.where(mask_stable)
                    .max("phase")
                    .assign_coords(phase=lambda x: phase[mask_stable.argmax("phase")])
                )

        if dim_to_suffix is not None:
            if isinstance(dim_to_suffix, str):
                dim_to_suffix = [dim_to_suffix]
            for dim in dim_to_suffix:
                out = out.pipe(dim_to_suffix_dataset, dim=dim)
        return out

    @property
    @xr_name(r"$\beta G(\mu,V,T)$", standard_name="Gibbs_free_energy")
    def betaG(self):
        r"""Scaled Gibbs free energy :math:`\beta G = \sum_i \beta \mu_i \overline{N}_i`."""
        # return self.betamu.pipe(lambda betamu: (betamu * self.nvec).sum(betamu.dims_comp))
        # return (self.betamu * self.nvec).sum(self.dims_comp)
        return xr.dot(
            self.betamu, self.nvec, dims=self.dims_comp, **self._xarray_dot_kws
        )

    @property
    @xr_name(r"$\beta G(\mu,V,T)/n$", standard_name="Gibbs_free_energy_per_particle")
    def betaG_n(self):
        r"""Scaled Gibbs free energy per particle :math:`\beta G / \overline{N}`."""
        return self.betaG / self.ntot

    @xr_name(r"$\beta F(\mu,V,T)$", standard_name="helmholtz_free_energy")
    def betaF(self, lnpi_zero=None):
        r"""Scaled Helmholtz free energy :math:`\beta F = \beta \Omega + \beta G`."""
        return self.betaOmega(lnpi_zero) + self.betaG

    @xr_name(
        r"$\beta F(\mu,V,T)/n$", standard_name="helmholtz_free_energy_per_particle"
    )
    def betaF_n(self, lnpi_zero=None):
        r"""Scaled Helmholtz free enenergy per particle :math:`\beta F / \overline{N}`."""
        return self.betaF(lnpi_zero) / self.ntot

    @xr_name(r"$\beta E(\mu,V,T)$", standard_name="total_energy")
    def betaE(self, ndim=3):
        r"""Scaled total energy :math:`\beta E = \beta \overline{PE} + \beta \overline{KE}`."""
        return ndim / 2.0 * self.ntot + self.PE.pipe(lambda x: x * x.beta)

    @xr_name(r"$\beta E(\mu,V,T)/ n$", standard_name="total_energy_per_particle")
    def betaE_n(self, ndim=3):
        r"""Scaled total energy per particle :math:`\beta E / \overline{N}`."""
        return self.betaE(ndim) / self.ntot

    @xr_name(r"$S(\mu,V,T) / k_{\rm B}$", standard_name="entropy")
    def S(self, lnpi_zero=None, ndim=3):
        r"""Scaled entropy :math:`S / k_{\rm B} = \beta E - \beta F`."""
        return self.betaE(ndim) - self.betaF(lnpi_zero)

    @xr_name(r"$S(\mu,V,T)/(n kB)$", standard_name="entropy_per_particle")
    def S_n(self, lnpi_zero=None, ndim=3):
        r"""Scaled entropy per particle :math:`S / (N k_{\rm B})`."""

        return self.S(lnpi_zero, ndim) / self.ntot


# NOTE: Using function accessor to override the docstring
@lnPiMasked.decorate_accessor("xce")
@MaskedlnPiDelayed.decorate_accessor("xce")
@MaskedlnPiLegacy.decorate_accessor("xce")
def xce_accessor(parent):
    """Accessor to :class:`~lnpy.xCanonical`"""
    return xCanonical(parent)


class xCanonical(object):
    """
    Canonical ensemble properties

    Parameters
    ----------
    parent : lnPiMasked
    """

    def __init__(self, parent):
        self._parent = parent
        self._xge = parent.xge

    def lnpi(self, fill_value=None):
        r"""DataArray view of :math:`\ln Pi(N)`"""
        return self._xge.lnpi(fill_value=fill_value)

    @property
    def _xarray_unstack(self):
        return getattr(self._parent, "_xarray_unstack", True)

    @gcached()
    def ncoords(self):
        """Coordinate vector `dims_n`"""
        return (
            self._xge.ncoords
            # NOTE: if don't use '.values', then get extra coords don't want
            .where(~self._xge.lnpi(np.nan).isnull().values)
        )

    @property
    def nvec(self):
        r"""Number of particles for each components :math:`{\bf N}`"""
        return self.ncoords.rename("nvec")

    @gcached()
    def ntot(self):
        """Total number of particles :math:`N`"""
        return self.ncoords.sum(self._xge.dims_comp)

    @gcached(prop=False)
    @xr_name(r"$\beta F({\bf n},V,T)$", standard_name="helmholtz_free_energy")
    def _betaF(self, lnpi_zero=None):
        """
        Helmholtz free energy
        """
        x = self._parent.xge

        if lnpi_zero is None:
            lnpi_zero = self._parent.data.ravel()[0]
            # lnpi_zero = x.lnpi_zero

        return (
            (-(x.lnpi(np.nan) - lnpi_zero) + (x.ncoords * x.betamu).sum(x.dims_comp))
            .assign_coords(**x._wrapper.coords_n)
            .drop(x.dims_lnz)
            .assign_attrs(x._standard_attrs)
        )

    def betaF(self, lnpi_zero=None):
        r"""Scaled Helmoltz free energy :math:`\beta F`"""
        return self._betaF(lnpi_zero)

    @xr_name(
        r"$\beta F({\bf n},V,T)/n$", standard_name="helmholtz_free_energy_per_particle"
    )
    def betaF_n(self, lnpi_zero=None):
        r"""Scaled Helmoltz free energy per particle :math:`\beta F / N`"""
        return self.betaF(lnpi_zero) / self.ntot

    @gcached()
    @xr_name(r"${\rm PE}({\bf n},V,T)/n$", standard_name="potential_energy")
    def PE(self):
        """Internal Energy :math:`PE`"""
        # if betaPE available, use that:
        PE = self._parent.extra_kws.get("PE", None)
        if PE is None:
            raise AttributeError('must set "PE" in "extra_kws" of lnPiMasked')
        x = self._xge
        coords = dict(x._wrapper.coords_n, **self._parent.state_kws)
        return xr.DataArray(PE, dims=x.dims_n, coords=coords, attrs=x._standard_attrs)

    @property
    @xr_name(
        r"${\rm PE}({\bf n},V,T)/n$", standard_name="potential_energy_per_particle"
    )
    def PE_n(self):
        r"""Internal energy per particle :math:`PE / N`"""
        return self.PE / self.ntot

    @xr_name(r"$\beta E({\bf n},V,T)$")
    def betaE(self, ndim=3):
        r"""Scaled total energy :math:`\beta E = \beta PE + \beta KE`"""
        return ndim / 2 * self.ntot + self._xge.beta * self.PE

    @xr_name(r"$\beta E({\bf n},V,T)/n$")
    def betaE_n(self, ndim=3):
        """Scaled total energy per particle"""
        return self.betaE(ndim) / self.ntot

    @xr_name(r"$S({\bf n},V,T)/k_{\rm B}$")
    def S(self, ndim=3, lnpi_zero=None):
        r"""Entropy :math:`S / k_{\rm B}`"""
        return self.betaE(ndim) - self.betaF(lnpi_zero)

    @xr_name(r"$S({\bf n},V,T)/(n k_{rm B})$")
    def S_n(self, ndim=3, lnpi_zero=None):
        r"""Entropy per particle :math:`S / (N k_{\rm B})`"""
        return self.S(ndim, lnpi_zero) / self.ntot

    @gcached(prop=False)
    @xr_name(r"$\beta {\bf\mu}({bf n},V,T)$", standard_name="absolute_activity")
    def _betamu(self, lnpi_zero=None):
        return xr.concat(
            [self.betaF(lnpi_zero).differentiate(n) for n in self._xge.dims_n],
            dim=self._xge.dims_comp[0],
        ).assign_attrs(self._xge._standard_attrs)

    def betamu(self, lnpi_zero=None):
        r"""Scaled chemical potential :math:`\beta \mu`"""
        return self._betamu(lnpi_zero)

    @property
    @xr_name(r"${\bf \rho}({\bf n},V,T)$")
    def dens(self):
        r"""Density :math:`\rho = N / V`"""
        return self.ncoords / self._xge.volume

    @gcached(prop=False)
    @xr_name(r"$\beta\Omega({\bf n},V,T)$")
    def _betaOmega(self, lnpi_zero=None):
        """
        beta * Omega = betaF - lnz .dot. N
        """
        return self.betaF(lnpi_zero) - (self.betamu(lnpi_zero) * self.ncoords).sum(
            self._xge.dims_comp
        )

    def betaOmega(self, lnpi_zero=None):
        r"""Scaled Grand Potential :math:`\beta \Omega`"""
        return self._betaOmega(lnpi_zero)

    @xr_name(r"$\beta\Omega({\bf n},V,T)/n$")
    def betaOmega_n(self, lnpi_zero=None):
        r"""Scaled Grand potential per particle, :math:`\beta\Omega / N`"""
        return self.betaOmega(lnpi_zero) / self.ntot

    @xr_name(r"$\beta p({\bf n},V,T)V$")
    def betapV(self, lnpi_zero=None):
        r""":math:`\beta p V = -\beta \Omega`"""
        return -self.betaOmega(lnpi_zero)

    @xr_name(r"$\beta p({\bf n},V,T)/\rho$")
    def Z(self, lnpi_zero=None):
        r"""Compressibility :math:`Z = \beta p / \rho = pV / (k_{\rm B} T)`"""
        return -self.betaOmega_n(lnpi_zero)

    @xr_name(r"$p({\bf n},V,T)$")
    def pressure(self, lnpi_zero=None):
        r"""Pressure :math:`p = -\Omega / V`"""
        return self.betapV(lnpi_zero).pipe(lambda x: x / (x.beta * x.volume))

    def table(
        self,
        keys=None,
        default_keys=("betamu", "betapV", "PE_n", "betaF_n"),
        dim_to_suffix=None,
    ):
        """
        Create Dataset from attributes/methods of `self`

        Parameters
        ----------
        keys : sequence of str
            Names of attributes/methods to access
        default_keys : sequence of str
            Default keys to access.
        dim_to_suffix : str, sequence of str, optional
            If passed, convert dimenensions in `dim_to_suffix` from dimensions in output to suffixes to variable names


        Returns
        -------
        table : Dataset
        """

        out = []

        def _process_keys(x):
            if x is None:
                return []
            elif isinstance(x, str):
                return [x]
            else:
                return list(x)

        keys = _process_keys(keys) + _process_keys(default_keys)

        # loop over unique keys only, in order
        for key in dict.fromkeys(keys):
            try:
                v = getattr(self, key)
                if callable(v):
                    v = v()
                out.append(v)
            except Exception:
                pass

        out = xr.merge(out)

        if dim_to_suffix is not None:
            if isinstance(dim_to_suffix, str):
                dim_to_suffix = [dim_to_suffix]
            for dim in dim_to_suffix:
                out = out.pipe(dim_to_suffix_dataset, dim=dim)
        return out



"""
Set of helper utilities to work with single component system
"""

# from functools import partial

import numpy as np
from scipy.optimize import brentq

from .lnpiseries import lnPiCollection


def get_lnz_min(
    target,
    C,
    build_phases,
    phase_id=0,
    component=None,
    dlnz=0.5,
    dfac=1.0,
    ref=None,
    build_kws=None,
    ntry=20,
    solve_kws=None,
    full_output=True,
):
    """
    solve for target density

    Parameters
    ----------
    target : float
        target density of specified component
    C : CollectionPhases
        initial guess to work from.  This is assumed to be sorted in accending lnz order
    build_phases : callable
        funciton to build new phases objects
    phase_id : int, default=0
        phase id tag to consider.  If None, consider min over 'phases'
    component : int, optional
        component to consider.
        If component is None, use build_phases.index.
        If component=='None', then consider total density
    dlnz : float, default=0.5
        how to change the chemical potential if `C` doesn't already
        bracket solution
    ref : lnPiMasked, optional
        optional lnPiMasked object to pass to build_phases
    build_kws : dictionary, optional
        optional arguments to `build_phases`
    ntry : int, default=20
        number of times to attempted finding left/right bracketing objects
    solve_kws : dictionary, optional
        optional arguments to `scipy.optimize.brentq`
    full_output : bool, default=True
        if True, return solve parameters

    Returns
    -------
    phases : Phases object as solution lnz
    solution_parameters : optional
    """

    if phase_id is not None and phase_id not in C.index.get_level_values("phase"):
        raise ValueError("no phase {}".format(phase_id))

    lnz_idx = build_phases.index
    if isinstance(component, str) and component.lower() == "none":

        def getter_comp(x):
            return x.xge.dens_tot

    else:
        if component is None:
            component = lnz_idx

        def getter_comp(x):
            return x.xge.dens.sel(component=component)

    if phase_id is None:

        def getter(x):
            return getter_comp(x).min("phase")

    else:

        def getter(x):
            return getter_comp(x).sel(phase=phase_id)

    s = getter(C).to_series().dropna()
    # builder
    if build_kws is None:
        build_kws = {}

    # left bracket
    left = None
    ss = s[s < target]
    if len(ss) > 0:
        left = C.mloc[ss.index[[-1]]]
    else:
        new_lnz = C.mloc[s.index[[0]]]._get_lnz(lnz_idx)
        dlnz_left = dlnz
        for i in range(ntry):
            new_lnz -= dlnz_left
            p = build_phases(new_lnz, ref=ref, **build_kws)
            if phase_id in p._get_level("phase") and getter(p).values < target:
                left = p
                break
            dlnz_left *= dfac

    if left is None:
        raise RuntimeError("could not find left bounds")

    # right bracket
    right = None
    ss = s[s > target]
    if len(ss) > 0:
        right = C.mloc[ss.index[[0]]]
    else:
        new_lnz = C.mloc[s.index[[-1]]]._get_lnz(lnz_idx)
        dlnz_right = dlnz

        for i in range(ntry):
            new_lnz += dlnz_right
            p = build_phases(new_lnz, ref=ref, **build_kws)

            if phase_id not in p._get_level("phase"):
                # went to far
                new_lnz -= dlnz_right
                # reset to half dlnz
                dlnz_right = dlnz_right * 0.5
            elif getter(p).values > target:
                right = p
                break
            else:
                dlnz_right *= dfac

    if right is None:
        raise RuntimeError("could not find right bounds")

    def f(x):
        lnz_new = x
        p = build_phases(lnz_new, ref=ref, **build_kws)
        f.lnpi = p
        return getter(p).values - target

    a, b = sorted([x._get_lnz(lnz_idx) for x in [left, right]])
    if solve_kws is None:
        solve_kws = {}
    xx, r = brentq(f, a, b, full_output=True, **solve_kws)
    r.residual = f(xx)

    if full_output:
        return f.lnpi, r
    else:
        return f.lnpi


def get_lnz_max(
    edge_distance_min,
    build_phases,
    ref=None,
    lnz_start=None,
    C=None,
    dlnz=0.5,
    dfac=1.0,
    threshold_abs=1e-4,
    ntry=50,
    build_kws=None,
    full_output=True,
):
    """
    find max lnz by bisection
    """

    if build_kws is None:
        build_kws = {}
    if ref is None:
        ref = build_phases.phase_creator.ref

    lnz_idx = build_phases.index
    if lnz_start is None:
        lnz_start = ref.lnz[lnz_idx]

    # need left/right bounds
    # left is greatest lnz point with edge_distance > edge_distance_min
    # right is least lnz point with edge_distance < edge_distance_min
    left = right = None
    lnz_left = lnz_right = None
    n_left = n_right = None

    def getter(p):
        v = p.xge.edge_distance(ref)
        if not p._xarray_unstack:
            v = v.unstack(p._concat_dim)
        return v.min("phase")

    # if have C, try to use it
    if C is not None:
        # see if bound contained in C

        s = getter(C).to_series()

        # left
        ss = s[s > edge_distance_min]
        if len(ss) > 0:
            left = C.mloc[ss.index[[-1]]]
        else:
            lnz_left = C.zloc[[0]]._get_lnz(lnz_idx)

        # right
        ss = s[s < edge_distance_min]
        if len(ss) > 0:
            right = C.mloc[ss.index[[0]]]
        else:
            lnz_right = C.zloc[[-1]]._get_lnz(lnz_idx)

    # if left not set, try to find it
    if left is None:
        if lnz_left is None:
            lnz_left = lnz_start
        dlnz_loc = dlnz
        for i in range(ntry):
            lnz_left -= dlnz_loc
            p = build_phases(lnz_left, ref=ref, **build_kws)
            if getter(p).values >= edge_distance_min:
                left = p
                n_left = i
                break
            # grow dlnz
            dlnz_loc *= dfac

    if left is None:
        raise RuntimeError("could not find left bound")

    # if right not set, try to find it
    if right is None:
        if lnz_right is None:
            lnz_right = lnz_start

        dlnz_loc = dlnz
        for i in range(ntry):
            lnz_right += dlnz_loc
            p = build_phases(lnz_right, ref=ref, **build_kws)

            if getter(p).values < edge_distance_min:
                right = p
                n_right = i
                break
            # grow dlnz
            dlnz_loc *= dfac

    if right is None:
        raise RuntimeError("could not find right bound")

    # not do bisection
    P = [left, right]
    Y = [getter(x).values for x in P]

    for i in range(ntry):
        lnz = [x._get_lnz(lnz_idx) for x in P]
        delta = np.abs(lnz[1] - lnz[0])
        if delta < threshold_abs:
            break
        lnz_mid = 0.5 * (lnz[0] + lnz[1])
        mid = build_phases(lnz_mid, ref=ref, **build_kws)
        y_mid = getter(mid).values

        if y_mid >= edge_distance_min:
            index = 0
        else:
            index = 1

        P[index] = mid
        Y[index] = y_mid

    left, right = P
    if full_output:
        info = dict(ntry=i, ntry_left=n_left, ntry_right=n_right, precision=delta)
        return left, info
    else:
        return left


class _BaseLimit(object):
    _bound_side = None  # -1 for lower bound, +1 for upper bound

    def __init__(self, parent):
        self._parent = parent

    @property
    def parent(self):
        return self._parent

    def __call__(self, *args, **kwargs):
        raise NotImplementedError("to be implemented in subclass")

    @property
    def lnz_idx(self):
        return self._lnz_idx

    @lnz_idx.setter
    def lnz_idx(self, i):
        self._lnz_idx = i

    @property
    def access(self):
        return self._access

    @access.setter
    def access(self, val):
        if not isinstance(val, lnPiCollection):
            raise ValueError("access should be lnPiCollection")
        self._access = val

    def _get_bound(self, sign, access=None, lnz_idx=None):
        if access is None:
            access = self.access
        if lnz_idx is None:
            lnz_idx = self.lnz_idx
        if sign == self._bound_side:
            return access._get_lnz(lnz_idx)
        else:
            return sign * np.inf

    def get_bounds(self, access=None, lnz_idx=None):
        return np.array(
            [
                self._get_bound(sign=sign, access=access, lnz_idx=lnz_idx)
                for sign in [-1, +1]
            ]
        )

    @property
    def bounds(self):
        return self._bounds

    @bounds.setter
    def bounds(self, bounds):
        assert len(bounds) == 2
        self._bounds = np.array(bounds)

    def set_bounds(self, access=None, lnz_idx=None):
        self.bounds = self.get_bounds(access=access, lnz_idx=lnz_idx)

    @property
    def lb(self):
        return self.bounds[0]

    @property
    def ub(self):
        return self.bounds[1]

    @property
    def lnz_list(self):
        return self._lnz_list

    def get_lnz_list(self, lnz=None, lnz_idx=None):
        if lnz is None:
            lnz = self.access.iloc[0].lnz
        if lnz_idx is None:
            lnz_idx = self.lnz_idx
        lnz = list(lnz)
        lnz[lnz_idx] = None
        return lnz

    def set_lnz_list(self, lnz=None, lnz_idx=None):
        self._lnz_list = self.get_lnz_list(lnz, lnz_idx)


@lnPiCollection.decorate_accessor("bounds_lower", single_create=True)
class LowerBounds(_BaseLimit):

    _bound_side = -1

    def __call__(
        self,
        dens_min,
        build_phases,
        phase_id=0,
        component=None,
        dlnz=0.5,
        dfac=1.0,
        ref=None,
        build_kws=None,
        ntry=20,
        solve_kws=None,
        inplace=True,
        force=False,
    ):

        if hasattr(self, "_access") and not force:
            p = self.access
            info = self._info
            lnz_idx = self.lnz_idx
            bounds = self.bounds

        else:
            lnz_idx = build_phases.index
            p, info = get_lnz_min(
                target=dens_min,
                C=self._parent,
                build_phases=build_phases,
                phase_id=phase_id,
                component=component,
                dlnz=dlnz,
                dfac=1.0,
                ref=ref,
                build_kws=build_kws,
                ntry=ntry,
                solve_kws=solve_kws,
            )

            bounds = self.get_bounds(p, lnz_idx)

        if inplace:
            self.access = p
            self._info = info
            self.lnz_idx = lnz_idx
            self.bounds = bounds
            self.set_lnz_list()
        else:
            return bounds, p, lnz_idx, info


@lnPiCollection.decorate_accessor("bounds_upper", single_create=True)
class UpperBounds(_BaseLimit):
    _bound_side = +1

    def __call__(
        self,
        edge_distance_min,
        build_phases,
        ref=None,
        lnz_start=None,
        dlnz=0.5,
        dfac=1.0,
        threshold_abs=1e-4,
        ntry=20,
        build_kws=None,
        inplace=True,
        force=False,
    ):

        if hasattr(self, "_access") and not force:
            p = self.access
            info = self._info
            lnz_idx = self.lnz_idx
            bounds = self.bounds

        else:
            lnz_idx = build_phases.index
            p, info = get_lnz_max(
                edge_distance_min=edge_distance_min,
                C=self._parent,
                build_phases=build_phases,
                lnz_start=None,
                dlnz=dlnz,
                dfac=1.0,
                threshold_abs=threshold_abs,
                ntry=ntry,
                ref=ref,
                build_kws=build_kws,
                full_output=True,
            )
            bounds = self.get_bounds(p, lnz_idx)

        if inplace:
            self.access = p
            self._info = info
            self.lnz_idx = lnz_idx
            self.bounds = bounds
            self.set_lnz_list()
        else:
            return bounds, p, lnz_idx, info


def build_grid(
    x=None,
    dx=None,
    x_range=None,
    x0=None,
    offsets=None,
    even_grid=False,
    digits=None,
    unique=True,
    outlier=False,
):
    """
    build a grid of values


    Parameters
    ----------
    x : array, optional
        if passes, use this as the base array
    dx : float, optional
        delta x. See below for uses
    x_range : array of length 2, optional
        if x not specified, make an array with
        `x=np.arange(x_range[0], x_range[1] + dx/2, dx)`
    x0 : float, optional
        see offsets
    offsets : array of length 2, optional
        if x or x_range not specified, set x_range = x0 + offsets
    even_grid : bool, default=True
        if True, make sure all values are multiples of dx
    digits : int, optional
        if passes, round output to this many digits
    outlier : bool, default=False
        if outlier is True, allow values outsize x

    Returns
    -------
    out : array of values
    """

    if x is None:
        assert dx is not None

        if x_range is None:
            x_range = x0 + np.array(offsets)
        assert len(x_range) == 2
        x = np.arange(x_range[0], x_range[1] + dx * 0.5, dx)

    if not outlier:
        lb, ub = x.min(), x.max()

    if even_grid:
        x = np.round(x / dx) * dx

    if digits is not None:
        x = np.round(x, digits)

    if unique:
        x = np.unique(x)

    if not outlier:
        x = x[(lb <= x) & (x <= ub)]

    return x


def limited_collection(
    build_phases,
    dlnz,
    lnz_range=None,
    offsets=None,
    digits=None,
    even_grid=True,
    course_step=None,
    edge_distance_min=None,
    dens_min=None,
    lnz_min_kws=None,
    lnz_max_kws=None,
    ref=None,
    build_kws=None,
    build_stability_kws=None,
    nmax=None,
    xarray_output=True,
    collection_kws=None,
    limit_course=False,
):
    """
    build a lnPiCollection over a range of lnz values

    Parameters
    ----------
    build_phases :
    """

    if lnz_range is None:
        if ref is None:
            _ref = build_phases.phase_creator.ref
        else:
            _ref = ref
        lnz_range = _ref.lnz[build_phases.index] + np.array(offsets)

    lnz_range = np.array(lnz_range)
    if even_grid:
        new_range = np.round(lnz_range / dlnz) * dlnz
        if new_range[0] < lnz_range[0]:
            new_range[0] += dlnz
        if new_range[1] > lnz_range[1]:
            new_range[1] -= dlnz
        lnz_range = new_range
    lnzs = np.arange(lnz_range[0], lnz_range[1] + dlnz * 0.5, dlnz)
    if even_grid:
        lnzs = np.round(lnzs / dlnz) * dlnz
    if digits is not None:
        lnzs = np.round(lnzs, digits)

    if course_step is None:
        course_step = max(len(lnzs) // 40, 1)
    if collection_kws is None:
        collection_kws = {}

    # get_collection = partial(
    #     lnPiCollection.from_builder,
    #     build_phases=build_phases,
    #     build_kws=build_kws, nmax=nmax, xarray_output=xarray_output,
    #     **collection_kws
    # )

    # limit lnz
    c_course = None
    lnz_min, lnz_max = lnzs[0], lnzs[-1]
    if edge_distance_min is not None or dens_min is not None:
        # c_course = get_collection(lnzs[::course_step])
        c_course = lnPiCollection.from_builder(
            lnzs[::course_step],
            build_phases=build_phases,
            build_kws=build_kws,
            nmax=nmax,
            xarray_output=xarray_output,
            **collection_kws,
        )

        if dens_min is not None:
            try:
                if lnz_min_kws is None:
                    lnz_min_kws = {}
                p_min, o = get_lnz_min(
                    dens_min, c_course, build_phases, build_kws=build_kws, **lnz_min_kws
                )
                if o.converged:
                    lnz_min = p_min.iloc[0].lnz[build_phases.index] - dlnz
            except Exception:
                pass

        if edge_distance_min is not None:
            try:
                if lnz_max_kws is None:
                    lnz_max_kws = {}
                p_max, o = get_lnz_max(
                    edge_distance_min,
                    build_phases,
                    build_kws=build_kws,
                    C=c_course,
                    **lnz_max_kws,
                )
                lnz_max = p_max.iloc[0].lnz[build_phases.index]
            except Exception:
                pass

    lnzs = lnzs[(lnzs >= lnz_min) & (lnzs <= lnz_max)]
    c = lnPiCollection.from_builder(
        lnzs,
        build_phases=build_phases,
        build_kws=build_kws,
        nmax=nmax,
        xarray_output=xarray_output,
        **collection_kws,
    )

    if c_course is None:
        c_course = c
    elif limit_course:
        slnz = "lnz_{}".format(build_phases.index)
        q = f"{slnz} >= {lnz_min} and {slnz} <= {lnz_max}"
        c_course = c_course.query(q)

    return c_course, c



from __future__ import absolute_import

################################################################################
# Delayed
from functools import lru_cache

import numpy as np
import pandas as pd

from ._docstrings import docfiller_shared
from .cached_decorators import gcached
from .extensions import AccessorMixin
from .utils import labels_to_masks, masks_change_convention

# from scipy.ndimage import filters


@lru_cache(maxsize=20)
def _get_n_ranges(shape, dtype):
    return [np.arange(s, dtype=dtype) for s in shape]


@lru_cache(maxsize=20)
def _get_shift(shape, dlnz, dtype):
    shift = np.zeros([], dtype=dtype)
    for i, (nr, m) in enumerate(zip(_get_n_ranges(shape, dtype), dlnz)):
        shift = np.add.outer(shift, nr * m)
    return shift


@lru_cache(maxsize=20)
def _get_data(base, dlnz):
    if all((x == 0 for x in dlnz)):
        return base._data
    else:
        return _get_shift(base.shape, dlnz, base._data.dtype) + base._data


@lru_cache(maxsize=20)
def _get_maskedarray(base, self, dlnz):
    return np.ma.MaskedArray(
        _get_data(base, dlnz), mask=self._mask, fill_value=base._fill_value
    )


@lru_cache(maxsize=20)
def _get_filled(base, self, dlnz, fill_value=None):
    return _get_maskedarray(base, self, dlnz).filled(fill_value)


class lnPiArray:
    """
    Wrapper on lnPi lnPiArray

    Parameters
    ----------
    lnz : float or sequence of floats
    """

    @docfiller_shared
    def __init__(
        self,
        lnz,
        data,
        state_kws=None,
        extra_kws=None,
        fill_value=np.nan,
        copy=False,
    ):
        """
        Parameters
        ----------
        {lnz}
        {data}
        {state_kws}
        {extra_kws}
        {fill_value}
        {copy}
        """

        lnz = np.atleast_1d(lnz)
        data = np.array(data, copy=copy)
        assert data.ndim == len(lnz)

        if state_kws is None:
            state_kws = {}
        if extra_kws is None:
            extra_kws = {}

        self._data = data
        # make data read only
        self._data.flags.writeable = False

        self._state_kws = state_kws
        self._extra_kws = extra_kws

        self._lnz = lnz
        self._fill_value = fill_value

    @property
    def shape(self):
        return self._data.shape

    def new_like(self, lnz=None, data=None, copy=False):
        """
        Create new object with optional replacements.

        All parameters are optional.  If not passed, use values in `self`

        Parameters
        ----------
        {lnz}
        {data}
        {copy}

        Returns
        -------
        out : lnPiArray
            New object with optionally updated parameters.

        """
        if lnz is None:
            lnz = self._lnz
        if data is None:
            data = self._data

        return type(self)(
            lnz=lnz,
            data=data,
            copy=copy,
            state_kws=self._state_kws,
            extra_kws=self._extra_kws,
            fill_value=self._fill_value,
        )

    def pad(self, axes=None, ffill=True, bfill=False, limit=None):
        """
        pad nan values in underlying data to values

        Parameters
        ----------
        ffill : bool, default=True
            Do forward filling
        bfill : bool, default=False
            Do back filling
        limit : int, default None
            The maximum number of consecutive NaN values to forward fill.
            If there is a gap with more than this number of
            consecutive NaNs, it will only be partially filled. Must be greater
            than 0 or None for no limit.
        Returns
        -------
        out : lnPiArray
            object with padded data
        """

        import bottleneck

        from .utils import bfill, ffill

        if axes is None:
            axes = range(self._data.ndim)

        data = self._data
        datas = []

        if ffill:
            datas += [ffill(data, axis=axis, limit=limit) for axis in axes]
        if bfill:
            datas += [bfill(data, axis=axis, limit=limit) for axis in axes]

        if len(datas) > 0:
            data = bottleneck.nanmean(datas, axis=0)

        new = self.new_like(data=data)
        return new

    def zeromax(self, mask=False):
        """
        shift values such that lnpi.max() == 0

        Parameters
        ----------
        mask : bool or array-like of bool
            Optional mask to apply to data.  Where `mask` is True,
            data is exluded from calculating maximum.
        """

        data = self._data - np.ma.MaskedArray(self._data, mask).max()
        return self.new_like(data=data)


@docfiller_shared
class lnPiMasked(AccessorMixin):
    """
    Masked array like wrapper for lnPi data.

    This is the basic data structure for storing the output from a single TMMC simulation.

    Parameters
    ----------
    {lnz}
    {base}
    {mask_masked}
    {copy}

    Notes
    -----
    Note that in most cases, :class:`lnPiMasked` should not be called directly.
    Rather, a constructor like :meth:`from_data` should be used to construct the object.

    Note the the value of `lnz` is the value to reweight to.

    Basic terminology:

    * T : temperature.
    * k : Boltzmanns constant.
    * beta : Inverse temperature `= 1/(k T)`.
    * mu : chemical potential.
    * lnz : log of activity `= ln(z)`.
    * z : activity `= beta * mu`.
    * lnPi : log of macrostate distribution.

    See Also
    --------
    numpy.ma.MaskedArray
    """

    _DataClass = lnPiArray

    def __init__(self, lnz, base, mask=None, copy=False):
        """ """
        lnz = np.atleast_1d(lnz)
        assert lnz.shape == base._lnz.shape

        if mask is None:
            mask = np.full(base._data.shape, fill_value=False, dtype=bool)
        else:
            mask = np.array(mask, copy=copy, dtype=bool)
        assert mask.shape == base._data.shape

        self._mask = mask
        # make mask read-only
        self._mask.flags.writeable = False

        self._base = base
        self._lnz = lnz
        self._dlnz = tuple(self._lnz - self._base._lnz)

    @classmethod
    def from_data(
        cls,
        lnz,
        lnz_data,
        data,
        mask=None,
        state_kws=None,
        extra_kws=None,
        fill_value=np.nan,
        copy=False,
    ):
        """
        Create :class:`lnPiMasked` object from raw data.

        Parameters
        ----------
        lnz : float or sequence of floats
            Value of `lnz` to reweight data to.
        lnz_data : float or sequence of floats
            Value of `lnz` at which `data` was collected
        {data}
        {mask_masked}
        {state_kws}
        {extra_kws}
        {fill_value}
        {copy}

        Returns
        -------
        out : lnPiMasked
        """

        base = cls._DataClass(
            lnz=lnz_data,
            data=data,
            state_kws=state_kws,
            extra_kws=extra_kws,
            fill_value=fill_value,
            copy=copy,
        )
        return cls(lnz=lnz, base=base, mask=mask, copy=copy)

    @property
    def _data(self):
        return self._base._data

    @property
    def dtype(self):
        """dtype of underling data"""
        return self._data.dtype

    def _clear_cache(self):
        self._cache = {}

    @property
    def state_kws(self):
        """'State' variables."""
        return self._base._state_kws

    @property
    def extra_kws(self):
        """'Extra' parameters."""
        return self._base._extra_kws

    @property
    def ma(self):
        """Masked array view of data reweighted data"""
        return _get_maskedarray(self._base, self, self._dlnz)

    def filled(self, fill_value=None):
        """Filled view or reweighted data"""
        return _get_filled(self._base, self, self._dlnz, fill_value)

    @property
    def data(self):
        """Reweighted data"""
        return _get_data(self._base, self._dlnz)

    @property
    def mask(self):
        """Where `True`, values are masked out."""
        return self._mask

    @property
    def shape(self):
        """lnPiArray shape"""
        return self._data.shape

    def __len__(self):
        return len(self._data)

    @property
    def ndim(self):
        return self._data.ndim

    @property
    def lnz(self):
        """Value of log(activity) evaluated at"""
        return self._lnz

    @property
    def betamu(self):
        """Alias to `self.lnz`"""
        return self._lnz

    @property
    def volume(self):
        """Accessor to self.state_kws['volume']."""
        return self.state_kws.get("volume", None)

    @property
    def beta(self):
        """Accessor to self.state_kws['beta']."""
        return self.state_kws.get("beta", None)

    def __repr__(self):
        return "<lnPi(lnz={})>".format(self._lnz)

    def __str__(self):
        return repr(self)

    def _index_dict(self, phase=None):
        out = {"lnz_{}".format(i): v for i, v in enumerate(self.lnz)}
        if phase is not None:
            out["phase"] = phase
        # out.update(**self.state_kws)
        return out

    # Parameters for xlnPi
    def _lnpi_tot(self, fill_value=None):
        return self.filled(fill_value)

    def _pi_params(self, fill_value=None):
        lnpi = self._lnpi_tot(fill_value)

        lnpi_local_max = lnpi.max()
        pi = np.exp(lnpi - lnpi_local_max)
        pi_sum = pi.sum()
        pi_norm = pi / pi_sum

        lnpi_zero = self.data.ravel()[0] - lnpi_local_max

        return pi_norm, pi_sum, lnpi_zero

    @property
    def _lnz_tot(self):
        return self.lnz

    # @gcached(prop=False)
    def local_argmax(self, *args, **kwargs):
        """Calculate index of maximum of masked data.

        Parameters
        ----------
        *args
            Positional arguments to argmax
        **kwargs
            Keyword arguments to argmax

        See Also
        --------
        numpy.ma.MaskedArray.argmax
        numpy.unravel_index
        """
        return np.unravel_index(self.ma.argmax(*args, **kwargs), self.shape)

    # @gcached(prop=False)
    def local_max(self, *args, **kwargs):
        """Calculate index of maximu of masked data.

        Parameters
        ----------
        *args
            Positional arguments to argmax
        **kwargs
            Keyword arguments to argmax

        See Also
        --------
        numpy.ma.MaskedArray.max
        """
        return self.ma[self.local_argmax(*args, **kwargs)]

    # @gcached(prop=False)
    def local_maxmask(self, *args, **kwargs):
        """Calculate mask where ``self.ma == self.local_max()``"""
        return self.ma == self.local_max(*args, **kwargs)

    @gcached()
    def edge_distance_matrix(self):
        """Matrix of distance from each element to a background (i.e., masked) point.

        See Also
        --------
        lnpy.utils.distance_matrix
        """
        from .utils import distance_matrix

        return distance_matrix(~self.mask)

    def edge_distance(self, ref, *args, **kwargs):
        """Distance of local maximum value to nearest background point.

        If `edge_distance` is too small, the value of properties calculated from this
        lnPi cannot be trusted.   This usually is due to the data being reweighted to
        too high a value of `lnz`, or not sampled to sufficiently high values of `N`.


        See Also
        --------
        edge_distance_matrix
        lnpy.utils.distance_matrix
        """
        return ref.edge_distance_matrix[self.local_argmax(*args, **kwargs)]

    @docfiller_shared
    def new_like(self, lnz=None, base=None, mask=None, copy=False):
        """
        Create new object with optional parameters

        Parameters
        ----------
        {lnz}
        {base}
        {mask_masked}
        {copy}
        """

        if lnz is None:
            lnz = self._lnz
        if base is None:
            base = self._base
        if mask is None:
            mask = self._mask

        return type(self)(lnz=lnz, base=base, mask=mask, copy=copy)

    def pad(self, axes=None, ffill=True, bfill=False, limit=None):
        """
        pad nan values in underlying data to values

        Parameters
        ----------
        ffill : bool, default=True
            do forward filling
        bfill : bool, default=False
            do back filling
        limit : int, default None
            The maximum number of consecutive NaN values to forward fill. In
            other words, if there is a gap with more than this number of
            consecutive NaNs, it will only be partially filled. Must be greater
            than 0 or None for no limit.
        inplace : bool, default=False

        Returns
        -------
        out : lnPi
            padded object


        See Also
        --------
        lnPiArray.pad
        """

        base = self._base.pad(axes=axes, ffill=ffill, bfill=bfill, limit=limit)
        return self.new_like(base=base)

    def zeromax(self):
        """
        shift so that lnpi.max() == 0 on reference

        See Also
        --------
        lnPiArray.zeromax
        """

        base = self._base.zeromax(mask=self._mask)
        return self.new_like(base=base)

    def reweight(self, lnz):
        """
        Create new object at specified value of `lnz`
        """
        return self.new_like(lnz=lnz)

    def or_mask(self, mask, **kwargs):
        """
        new object with logical or of self.mask and mask
        """
        return self.new_like(mask=(mask | self.mask))

    def and_mask(self, mask, **kwargs):
        """
        new object with logical and of self.mask and mask
        """
        return self.copy_shallow(mask=(mask & self.mask))

    @classmethod
    def from_table(
        cls, path, lnz, state_kws=None, sep=r"\s+", names=None, csv_kws=None, **kwargs
    ):
        """
        Create lnPi object from text file table with columns [n_0,...,n_ndim, lnpi]

        Parameters
        ----------
        path : string like
            file object to be read
        lnz : array-like
            beta*(chemical potential) for each component
        state_kws : dict, optional
            define state variables, like volume, beta
        sep : string, optional
            separator for file read
        names : column names
        csv_kws : dict, optional
            optional arguments to `pandas.read_csv`
        kwargs  : extra arguments
            Passed to lnPi constructor
        """
        lnz = np.atleast_1d(lnz)
        ndim = len(lnz)

        if names is None:
            names = ["n_{}".format(i) for i in range(ndim)] + ["lnpi"]

        if csv_kws is None:
            csv_kws = {}

        da = (
            pd.read_csv(path, sep=sep, names=names, **csv_kws)
            .set_index(names[:-1])["lnpi"]
            .to_xarray()
        )
        return cls.from_data(
            data=da.values,
            mask=da.isnull().values,
            lnz=lnz,
            lnz_data=lnz,
            state_kws=state_kws,
            **kwargs
        )

    @classmethod
    def from_dataarray(cls, da, state_as_attrs=None, **kwargs):
        """
        create a lnPi object from :class:`xarray.DataArray`

        Parameters
        ----------
        da : DataArray
            DataArray containing the lnPi data
        state_as_attrs : bool, optional
            If True, get `state_kws` from ``da.attrs``.
        **kwargs
            Extra arguments to :meth:`from_data`

        Returns
        -------
        lnpi : lnPiMasked instance


        See Also
        --------
        :meth:`from_data`

        """

        kws = {}
        kws["data"] = da.values
        if "mask" in da.coords:
            kws["mask"] = da.mask.values
        else:
            kws["mask"] = da.isnull().values

        # where are state variables
        if state_as_attrs is None:
            state_as_attrs = bool(da.attrs.get("state_as_attrs", False))
        if state_as_attrs:
            # state variables from attrs
            c = da.attrs
        else:
            c = da.coords

        lnz = []
        state_kws = {}
        for k in da.attrs["dims_state"]:
            val = np.array(c[k])
            if "lnz" in k:
                lnz.append(val)
            else:
                if val.ndim == 0:
                    val = val[()]
                state_kws[k] = val
        kws["lnz"] = lnz
        kws["state_kws"] = state_kws

        kws["lnz_data"] = lnz

        # any overrides
        kwargs = dict(kws, **kwargs)
        return cls.from_data(**kwargs)

    @docfiller_shared
    def list_from_masks(self, masks, convention="image"):
        """
        create list of lnpis corresponding to masks[i]

        Parameters
        ----------
        {masks_masked}
        {mask_convention}

        Returns
        -------
        lnpis : list
            list of lnpis corresponding to each mask

        See Also
        --------
        masks_change_convention
        """

        return [
            self.or_mask(m) for m in masks_change_convention(masks, convention, False)
        ]

    @docfiller_shared
    def list_from_labels(
        self,
        labels,
        features=None,
        include_boundary=False,
        check_features=True,
        **kwargs
    ):
        """
        Create sequence of lnpis from labels array.

        Parameters
        ----------
        {labels}
        features : sequence of ints, optional
            If specified, extract only those locations where ``labels == feature``
            for all values ``feature in features``.  That is, select a subset of
            unique label values.
        include_boundary : bool, default=False
            if True, include boundary regions in output mask
        check_features : bool, default=True
            if True, and supply features, then make sure each feature is in labels
        **kwargs
            Extra arguments to to :func:`~lnpy.utils.labels_to_masks`

        Returns
        -------
        outputs : list of lnPiMasked objects

        See Also
        --------
        lnPiMasked.list_from_masks
        labels_to_masks
        """

        masks, features = labels_to_masks(
            labels=labels,
            features=features,
            include_boundary=include_boundary,
            convention=False,
            check_features=check_features,
            **kwargs
        )
        return self.list_from_masks(masks, convention=False)


from warnings import warn


class MaskedlnPiDelayed(lnPiMasked):
    def __init__(self, *args, **kwargs):
        warn("MaskedlnPiDelayed is deprecated.  Please use lnPiMasked instead")
        super().__init__(*args, **kwargs)



"""
utility functions
"""

from functools import partial

import numpy as np

# import pandas as pd
import xarray as xr
from scipy import ndimage as ndi
from skimage import segmentation

# --------------------------------------------------
# TQDM stuff
try:
    import tqdm as _tqdm

    _HAS_TQDM = True
except ImportError:
    _HAS_TQDM = False

if _HAS_TQDM:
    try:
        from IPython import get_ipython

        p = get_ipython()
        if p is not None and p.has_trait("kernel"):
            from tqdm.notebook import tqdm as tqdm_default
        else:
            tqdm_default = _tqdm.tqdm
    except ImportError:
        tqdm_default = _tqdm.tqdm

from .options import OPTIONS


def tqdm(*args, **kwargs):
    opt = OPTIONS["tqdm_bar"]
    if opt == "text":
        func = _tqdm.tqdm
    elif opt == "notebook":
        func = _tqdm.tqdm_notebook
    else:
        func = tqdm_default

    return func(*args, **kwargs)


def get_tqdm(seq, len_min, leave=None, **kwargs):

    n = kwargs.get("total", None)

    if isinstance(len_min, str):
        len_min = OPTIONS[len_min]

    if n is None:
        n = len(seq)
    if _HAS_TQDM and OPTIONS["tqdm_use"] and n >= len_min:
        if leave is None:
            leave = OPTIONS["tqdm_leave"]
        seq = tqdm(seq, leave=leave, **kwargs)
    return seq


get_tqdm_calc = partial(get_tqdm, len_min="tqdm_len_calc")
get_tqdm_build = partial(get_tqdm, len_min="tqdm_len_build")

# def get_tqdm_calc(seq, len_min=None, leave=None, **kwargs):
#     if len_min is None:
#         len_min = OPTIONS['tqdm_len_calc']
#     return _get_tqdm(seq, len_min=len_min, leave=leave, **kwargs)


# def get_tqdm_build(seq, len_min=None, leave=None, **kwargs):
#     if len_min is None:
#         len_min = OPTIONS['tqdm_len_build']
#     return _get_tqdm(seq, len_min=len_min, leave=leave, **kwargs)


# --------------------------------------------------
# JOBLIB stuff
try:
    from joblib import Parallel, delayed

    _HAS_JOBLIB = True
except ImportError:
    _HAS_JOBLIB = False


from operator import attrgetter


def parallel_map_build(func, items, *args, **kwargs):
    if (
        _HAS_JOBLIB
        and OPTIONS["joblib_use"]
        and len(items) >= OPTIONS["joblib_len_build"]
    ):
        return Parallel(
            n_jobs=OPTIONS["joblib_n_jobs"],
            backend=OPTIONS["joblib_backend"],
            **OPTIONS["joblib_kws"]
        )(delayed(func)(x, *args, **kwargs) for x in items)
    else:
        return [func(x, *args, **kwargs) for x in items]


def _func_call(x, *args, **kwargs):
    return x(*args, **kwargs)


def parallel_map_call(items, use_joblib, *args, **kwargs):
    if (
        use_joblib
        and _HAS_JOBLIB
        and OPTIONS["joblib_use"]
        and len(items) >= OPTIONS["joblib_len_calc"]
    ):
        return Parallel(
            n_jobs=OPTIONS["joblib_n_jobs"],
            backend=OPTIONS["joblib_backend"],
            **OPTIONS["joblib_kws"]
        )(delayed(_func_call)(x, *args, **kwargs) for x in items)
    else:
        return [x(*args, **kwargs) for x in items]


def parallel_map_attr(attr, use_joblib, items):
    func = attrgetter(attr)
    if (
        use_joblib
        and _HAS_JOBLIB
        and OPTIONS["joblib_use"]
        and len(items) >= OPTIONS["joblib_len_calc"]
    ):
        return Parallel(
            n_jobs=OPTIONS["joblib_n_jobs"],
            backend=OPTIONS["joblib_backend"],
            **OPTIONS["joblib_kws"]
        )(delayed(func)(x) for x in items)
    else:
        return [func(x) for x in items]


def parallel_map_func_starargs(func, use_joblib, items, total=None):

    if total is None:
        total = len(items)

    if (
        use_joblib
        and _HAS_JOBLIB
        and OPTIONS["joblib_use"]
        and total >= OPTIONS["joblib_len_calc"]
    ):
        return Parallel(
            n_jobs=OPTIONS["joblib_n_jobs"],
            backend=OPTIONS["joblib_backend"],
            **OPTIONS["joblib_kws"]
        )(delayed(func)(*x) for x in items)
    else:
        return [func(*x) for x in items]


# ----------------------------------------
# pandas stuff
def allbut(levels, *names):
    names = set(names)
    return [item for item in levels if item not in names]


# ----------------------------------------
# xarray utils
def dim_to_suffix_dataarray(da, dim, join="_"):
    if dim in da.dims:
        return da.assign_coords(
            **{dim: lambda x: ["{}{}{}".format(x.name, join, c) for c in x[dim].values]}
        ).to_dataset(dim=dim)
    else:
        return da.to_dataset()


def dim_to_suffix_dataset(table, dim, join="_"):
    out = table
    for k in out:
        if dim in out[k].dims:
            out = out.drop(k).update(table[k].pipe(dim_to_suffix_dataarray, dim, join))
    return out


def dim_to_suffix(ds, dim="component", join="_"):
    if isinstance(ds, xr.DataArray):
        f = dim_to_suffix_dataarray
    elif isinstance(ds, xr.Dataset):
        f = dim_to_suffix_dataset
    else:
        raise ValueError("ds must be `DataArray` or `Dataset`")
    return f(ds, dim=dim, join=join)


def _convention_to_bool(convention):
    if convention == "image":
        convention = True
    elif convention == "masked":
        convention = False
    else:
        assert convention in [True, False]
    return convention


def mask_change_convention(mask, convention_in="image", convention_out="masked"):
    """
    convert an array from one 'mask' convention to another.

    Parameters
    ----------
    mask : array-like
        Masking array.
    convention_in, convention_out : string or bool or None.
        Convention for input and output.
        Convention for mask.  Allowable values are:

        * 'image' or True : `True` values included, `False` values excluded.
          This is the normal convention in :mod:`scipy.ndimage`.
        * 'masked' or False:  `False` values are included, `True` values are excluded.
          This is the convention in :mod:`numpy.ma`

        If `None`, then pass return input `mask`

    Returns
    -------
    new_mask : array
    New 'mask' array with specified convension.
    """

    if mask is None:
        return mask

    convention_in = _convention_to_bool(convention_in)
    convention_out = _convention_to_bool(convention_out)

    if convention_in != convention_out:
        mask = ~mask
    return mask


def masks_change_convention(masks, convention_in="image", convention_out="masked"):
    """
    Perform convension change of sequence of masks

    Parameters
    ----------
    masks : sequence of array-like
        masks[i] is the 'ith' mask
    convention_in, convention_out : string or bool or None.
        Convention for input and output.
        Convention for mask.  Allowable values are:

        * 'image' or True : `True` values included, `False` values excluded.
          This is the normal convention in :mod:`scipy.ndimage`.
        * 'masked' or False:  `False` values are included, `True` values are excluded.
          This is the convention in :mod:`numpy.ma`

        If `None`, then pass return input `mask`

    Returns
    -------
    new_masks : list of arrays
        New 'masks' array with specified convension.
    """

    convention_in = _convention_to_bool(convention_in)
    convention_out = _convention_to_bool(convention_out)

    if convention_in != convention_out:
        masks = [m if m is None else ~m for m in masks]
    return masks


##################################################
# labels/masks utilities
##################################################
def labels_to_masks(
    labels,
    features=None,
    include_boundary=False,
    convention="image",
    check_features=True,
    **kwargs
):
    """
    convert labels array to list of masks

    Parameters
    ----------
    labels : array of labels to analyze
        Each unique value `i` in `labels` indicates a mask.
        That is ``labels == i``.
    features : array-like, optional
        list of features to extract from labels.  Note that returned
        mask[i] corresponds to labels == feature[i].
    include_boundary : bool, default=False
        if True, include boundary regions in output mask
    convention : {'image','masked'} or bool.
        convention for output masks
    check_features : bool, default=True
        if True, and supply features, then make sure each feature is in labels

    **kwargs
        arguments to find_boundary if include_boundary is True
        default to mode='outer', connectivity=labels.ndim

    Returns
    -------
    output : list of masks of same shape as labels
        mask for each feature
    features : list
        features

    See Also
    --------
    masks_to_labels
    :func:`skimage.segmentation.find_boundaries`

    """

    if include_boundary:
        kwargs = dict(dict(mode="outer", connectivity=labels.ndim), **kwargs)
    if features is None:
        features = [i for i in np.unique(labels) if i > 0]
    elif check_features:
        vals = np.unique(labels)
        assert np.all([x in vals for x in features])

    convention = _convention_to_bool(convention)

    output = []
    for i in features:
        m = labels == i
        if include_boundary:
            b = segmentation.find_boundaries(m.astype(int), **kwargs)
            m = m + b
        if not convention:
            m = ~m
        output.append(m)
    return output, features


def masks_to_labels(masks, features=None, convention="image", dtype=int, **kwargs):
    """
    convert list of masks to labels

    Parameters
    ----------
    masks : list-like of masks

    features : value for each feature, optional
        labels[mask[i]] = features[i] + feature_offset
        Default = range(1, len(masks) + 1)

    convention : {'image','masked'} or bool
        convention of masks

    Returns
    -------
    labels : array of labels


    See Also
    --------
    labels_to_masks
    """

    if features is None:
        features = range(1, len(masks) + 1)
    else:
        assert len(features) == len(masks)

    labels = np.full(masks[0].shape, fill_value=0, dtype=dtype)

    masks = masks_change_convention(masks, convention, True)

    for i, m in zip(features, masks):
        labels[m] = i
    return labels


def ffill(arr, axis=-1, limit=None):
    import bottleneck

    _limit = limit if limit is not None else arr.shape[axis]
    return bottleneck.push(arr, n=_limit, axis=axis)


def bfill(arr, axis=-1, limit=None):
    """inverse of ffill"""
    import bottleneck

    # work around for bottleneck 178
    _limit = limit if limit is not None else arr.shape[axis]

    arr = np.flip(arr, axis=axis)
    # fill
    arr = bottleneck.push(arr, axis=axis, n=_limit)
    # reverse back to original
    return np.flip(arr, axis=axis)


##################################################
# calculations
##################################################


def get_lnz_iter(lnz, x):
    """
    create a lnz_iter object for varying a single lnz

    Parameters
    ----------
    lnz : list
        list with one element equal to None.  This is the component which will be varied
        For example, lnz=[lnz0,None,lnz2] implies use values of lnz0,lnz2 for components 0 and 2, and
        vary component 1

    x : array
        values to insert for variable component

    Returns
    -------
    ouptut : array of shape (len(x),len(lnz))
       array with rows [lnz0,lnz1,lnz2]
    """

    z = np.zeros_like(x)

    x = np.asarray(x)

    L = []
    for m in lnz:
        if m is None:
            L.append(x)
        else:
            L.append(z + m)

    return np.array(L).T


##################################################
# utilities
##################################################


def sort_lnPis(input, comp=0):
    """
    sort list of lnPi  that component `comp` mol fraction increases

    Parameters
    ----------
    input : list of lnPi objects


    comp : int (Default 0)
     component to sort along

    Returns
    -------
    output : list of lnPi objects in sorted order
    """

    molfrac_comp = np.array([x.molfrac[comp] for x in input])

    order = np.argsort(molfrac_comp)

    output = [input[i] for i in order]

    return output


def distance_matrix(mask, convention="image"):
    """
    create matrix of distances from elements of mask
    to nearest background point

    Parameters
    ----------
    mask : array-like
        image mask
    convention : str or bool, default='image'
        mask convetion

    Returns
    -------
    distance : array of same shape as mask
        distance from possible feature elements to background


    See Also
    --------
    scipy.ndimage.distance_transform_edt
    """

    mask = np.asarray(mask, dtype=bool)
    mask = masks_change_convention(mask, convention_in=convention, convention_out=True)

    # pad mask
    # add padding to end of matrix in each dimension
    ndim = mask.ndim
    pad_width = ((0, 1),) * ndim
    mask = np.pad(mask, pad_width=pad_width, mode="constant", constant_values=False)

    # distance filter
    dist = ndi.distance_transform_edt(mask)

    # remove padding
    s = (slice(None, -1),) * ndim
    return dist[s]


def lnpimasked_to_dataset(data, keys=("lnpi", "PE")):
    """
    Convert a :class:`~lnpy.lnPiMasked` object into as :class:`~xarray.Dataset`.

    Parameters
    ----------
    data : lnPiMasked

    Returns
    -------
    output : Dataset
    """

    return data.xce.table(keys=keys, default_keys=None)


def dataset_to_lnpimasked(ds, lnpi_name="lnpi", pe_name="PE", extra_kws=None, **kwargs):
    """
    Convert a :class:`~xarray.Dataset` to a :class:`~lnpy.lnPiMasked` object.

    Parameters
    ----------
    ds : Dataset

    lnpi_name, pe_name : str
        Names of 'lnPi' and 'PE' parameters

    extra_kws : mapping, optional
        parameter `extra_kws`.  Note that if `pe_name` found in `ds`, then will add it to `extra_kws`.

    Returns
    -------
    lnpi : lnPiMasked
    """

    from .lnpidata import lnPiMasked

    data = ds[lnpi_name]

    if extra_kws is None:
        extra_kws = {}

    if pe_name in ds:
        extra_kws[pe_name] = ds[pe_name].values

    return lnPiMasked.from_dataarray(da=data, extra_kws=extra_kws, **kwargs)



"""Define accessor routines.

This is inspired by xarray accessors.
"""

import warnings
from itertools import chain
from operator import attrgetter

from .cached_decorators import gcached
from .utils import get_tqdm_calc as get_tqdm
from .utils import parallel_map_attr, parallel_map_call


class AccessorRegistrationWarning(Warning):
    """Warning for conflicts in accessor registration."""


class _CachedAccessorSingle(object):
    """
    Custom property-like object (descriptor).
    this gets created once on call
    """

    def __init__(self, name, accessor):

        self.__doc__ = accessor.__doc__

        self._name = name
        self._accessor = accessor

    def __get__(self, obj, cls):
        if obj is None:
            return self._accessor
        try:
            accessor_obj = self._accessor(obj)
        except AttributeError:
            # __getattr__ on data object will swallow any AttributeErrors
            # raised when initializing the accessor, so we need to raise as
            # something else (GH933):
            raise RuntimeError("error initializing %r accessor." % self._name)
        # Replace the property with the accessor object. Inspired by:
        # http://www.pydanny.com/cached-property.html
        # We need to use object.__setattr__ because we overwrite __setattr__ on
        # AttrAccessMixin.
        object.__setattr__(obj, self._name, accessor_obj)
        return accessor_obj


from functools import wraps


def _CachedAccessorCleared(name, accessor):
    """
    Wrap accessor in a cached property

    This creates an property in the parent class with name `name`
    from `accessor(self)`.
    This only gets initialized when called

    Notes
    -----
    This gets deleted if self._cache is cleared.
    If you only want to create it once, then use the Single access wrapper below
    """

    @gcached(key=name, prop=True)
    @wraps(accessor)
    def _get_prop(self):
        return accessor(self)

    return _get_prop


def _CachedAccessorWrapper(name, accessor, single_create=False):
    if single_create:
        return _CachedAccessorSingle(name, accessor)
    else:
        return _CachedAccessorCleared(name, accessor)


class AccessorMixin(object):
    @classmethod
    def _register_accessor(cls, name, accessor, single_create=False):
        """
        most general accessor
        """
        if hasattr(cls, name):
            warnings.warn(
                "registration of accessor %r under name %r for type %r is "
                "overriding a preexisting attribute with the same name."
                % (accessor, name, cls),
                AccessorRegistrationWarning,
                stacklevel=2,
            )
        setattr(cls, name, _CachedAccessorWrapper(name, accessor, single_create))

    @classmethod
    def register_accessor_flat(cls, names, single_create=True, subattr="flat"):
        """
        helper class to register stuff to top of CollectionPhases
        """

        if isinstance(names, str):
            names = [names]
        for name in names:
            cls._register_accessor(
                name=name,
                accessor=attrgetter(".".join((subattr, name))),
                single_create=single_create,
            )

    @classmethod
    def register_accessor(cls, name, accessor, single_create=False):
        """
        Register a property `name` to `class` of type `accessor(self)`

        Examples
        --------
        >>> class parent(AccessorMixin):
        ...     pass
        ...
        >>> class hello(AccessorMixin):
        ...     def __init__(self, parent):
        ...         self._parent = parent
        ...
        ...     def there(self):
        ...         return f"{type(self._parent)}"
        ...

        >>> parent.register_accessor("hello", hello)
        >>> x = parent()
        >>> x.hello.there()
        'parent'
        """
        cls._register_accessor(name, accessor, single_create)

    @classmethod
    def decorate_accessor(cls, name, single_create=False):
        """
        Register a property `name` to `class` of type `accessor(self)`.

        Examples
        --------
        >>> class parent(AccessorMixin):
        ...     pass
        ...

        >>> @parent.decorate("hello")
        ... class hello(AccessorMixin):
        ...     def __init__(self, parent):
        ...         self._parent = parent
        ...
        ...     def there(self):
        ...         return f"{type(self._parent)}"

        >>> x = parent()
        >>> x.hello.there()
        'parent'
        """

        def decorator(accessor):
            cls.register_accessor(name, accessor, single_create)
            return accessor

        return decorator


################################################################################
# List access


class _CallableListResultsCache(object):
    """
    if items of collection accessor are callable, then
    """

    def __init__(self, parent, items, desc=None):
        self.parent = parent
        self.items = items
        self.desc = desc
        self._cache = {}
        self._use_joblib = getattr(self.parent, "_USE_JOBLIB_", False)

    @gcached(prop=False)
    def __call__(self, *args, **kwargs):
        # get value
        seq = get_tqdm(self.items, desc=self.desc)
        # results = [x(*args, **kwargs) for x in seq]
        results = parallel_map_call(seq, self._use_joblib, *args, **kwargs)
        if hasattr(self.parent, "wrap_list_results"):
            results = self.parent.wrap_list_results(results)
        return results


class _CallableListResultsNoCache(object):
    """
    if items of collection accessor are callable, then
    """

    def __init__(self, parent, items, desc=None):
        self.parent = parent
        self.items = items
        self.desc = desc
        self._cache = {}
        self._use_joblib = getattr(self.parent, "_USE_JOBLIB_", False)

    def __call__(self, *args, **kwargs):
        # get value
        seq = get_tqdm(self.items, desc=self.desc)
        # results = [x(*args, **kwargs) for x in seq]
        results = parallel_map_call(seq, self._use_joblib, *args, **kwargs)
        if hasattr(self.parent, "wrap_list_results"):
            results = self.parent.wrap_list_results(results)
        return results


def _CallableListResults(parent, items, use_cache=False, desc=None):
    if use_cache:
        cls = _CallableListResultsCache
    else:
        cls = _CallableListResultsNoCache

    return cls(parent=parent, items=items, desc=desc)


class _ListAccessor(object):
    def __init__(self, parent, items, cache_list=None):
        self.parent = parent
        self.items = items

        if cache_list is None:
            cache_list = []
        self._cache_list = cache_list
        self._cache = {}
        self._use_joblib = getattr(self.parent, "_USE_JOBLIB_", False)

    def __getattr__(self, attr):
        if attr in self._cache:
            # print('using cache')
            return self._cache[attr]

        try:
            use_cache = attr in self._cache_list
            if callable(getattr(self.items[0], attr)):
                seq = [getattr(x, attr) for x in self.items]
                result = _CallableListResults(
                    self.parent, seq, use_cache=use_cache, desc=attr
                )
            else:
                seq = get_tqdm(self.items, desc=attr)
                # result = [getattr(x, attr) for x in seq]
                result = parallel_map_attr(attr, items=seq, use_joblib=self._use_joblib)
                if hasattr(self.parent, "wrap_list_results"):
                    result = self.parent.wrap_list_results(result)

        except Exception:
            raise AttributeError(f"no attribute {attr} found")

        # do caching?
        if use_cache:
            # print('caching')
            self._cache[attr] = result
        return result

    def __getitem__(self, idx):
        return self.items[idx]

    def __dir__(self):
        heritage = dir(super(self.__class__, self))
        # hide = []
        x = self.items[0]
        show = [
            k
            for k in chain(
                self.__dict__.keys(),
                dir(x)
                # self.__class__.__dict__.keys()
                # x.__dict__.keys(),
                # x.__class__.__dict__.keys()
            )
        ]  # if k not in hide]
        return sorted(heritage + show)


def _CachedListPropertyWrapper(name, cache_list=None):
    """
    get top level property from items
    """

    if cache_list is not None and name in cache_list:
        cache = True
    else:
        cache = False

    # @gcached(key=name, prop=True)
    def _get_prop(self):
        results = [getattr(x, name) for x in self]
        if callable(results[0]):
            results = _CallableListResults(self, results)
        else:
            if hasattr(self, "wrap_list_results"):
                results = self.wrap_list_results(results)
        return results

    if cache:
        _get_prop = gcached(key=name, prop=True)(_get_prop)
    else:
        _get_prop = property(_get_prop)

    return _get_prop


def _CachedListAccessorWrapper(name, cache_list=None):
    """
    Wrap List accessor in cached property
    """

    @gcached(key=name, prop=True)
    def _get_prop(self):
        return _ListAccessor(
            self, [getattr(x, name) for x in self], cache_list=cache_list
        )

    return _get_prop


class ListAccessorMixin(object):
    @classmethod
    def _register_listaccessor(cls, names, accessor_wrapper, **kwargs):
        """
        most general accessor
        """
        if isinstance(names, str):
            names = [names]
        for name in names:
            if hasattr(cls, name):
                warnings.warn(
                    "registration of name %r for type %r is "
                    "overriding a preexisting attribute with the same name."
                    % (name, cls),
                    AccessorRegistrationWarning,
                    stacklevel=2,
                )
            setattr(cls, name, accessor_wrapper(name, **kwargs))

    @classmethod
    def register_listaccessor(cls, names, cache_list=None):
        """
        creat an accessor to elements of parent class.

        Examples
        --------
        class Prop(object):
            def __init__(self, prop)
                self.prop = prop
        class Child(object):
            def __init__(self, a, b):
                self.a = Prop(a)
        class Wrapper(ListAccessorMixin):
            def __init__(self, items):
                self.items = items
            def __getitem__(self, idx):
                # must have getitem for this to work
                return self.items[idx]
            def wrap_list_results(self, result):
                #anython special to do with resultsgoes here
                return results

        wrapper.register_listaccessor('a')
        >>> x = [Child(i) for i in range(3)]
        >>> w = Wrapper(x)
        >>> w.a.prop
        [1, 2, 3]
        """
        return cls._register_listaccessor(
            names, accessor_wrapper=_CachedListAccessorWrapper, cache_list=cache_list
        )

    @classmethod
    def register_listproperty(cls, names, cache_list=None):
        """
        create an accessor to property

        This is to give top level access to stuff

        Examples
        --------
        class Child(object):
            def __init__(self, a):
                self.a = a
            def b(self, x):
                return self.a, x
        class Wrapper(ListAccessorMixin):
            def __init__(self, items):
                self.items = items

            def __getitem__(self, idx):
                # must have getitem for this to work
                return self.items[idx]

            def wrap_list_results(self, result):
                #anython special to do with resultsgoes here
                return results

        Wrapper.register_listproperty(['a', 'b'])
        >>> x = [Child(i) for i in range(3)]
        >>> w = Wrapper(x)
        >>> w.a
        [1, 2, 3]
        >>> w.b("a")
        [(1,'a'), (2,'a'), (3,'a')]
        """
        return cls._register_listaccessor(
            names, accessor_wrapper=_CachedListPropertyWrapper, cache_list=cache_list
        )


def _decorate_listaccessor(names, accessor_wrapper, **kwargs):
    if isinstance(names, str):
        names = [names]

    def decorator(cls):
        for name in names:
            if hasattr(cls, name):
                warnings.warn(
                    "registration of name %r for type %r is "
                    "overriding a preexisting attribute with the same name."
                    % (name, cls),
                    AccessorRegistrationWarning,
                    stacklevel=2,
                )
            setattr(cls, name, accessor_wrapper(name, **kwargs))
        return cls

    return decorator


def decorate_listaccessor(names, cache_list=None):
    return _decorate_listaccessor(
        names, _CachedListAccessorWrapper, cache_list=cache_list
    )


def decorate_listproperty(names, cache_list=None):
    return _decorate_listaccessor(
        names, _CachedListPropertyWrapper, cache_list=cache_list
    )



import itertools
import warnings

import numpy as np
import pandas as pd
import xarray as xr
from skimage import segmentation

from ._docstrings import docfiller_shared
from .cached_decorators import gcached
from .lnpiseries import lnPiCollection
from .utils import get_tqdm_calc as get_tqdm
from .utils import labels_to_masks, masks_change_convention, parallel_map_func_starargs


def find_boundaries(masks, mode="thick", connectivity=None, **kws):
    """
    find boundary region for masks

    Parameters
    ----------
    masks : list of arrays of bool
        Convesions is `masks[i][index] == True`` if `index` is active for the ith mask
    mode : str, default="thick"
        mode to use in :func:`skimage.segmentation.find_boundaries`
    connectivity : int, default = masks[0].ndim


    Returns
    -------
    boundaries : list of arrays of bool, optional
        If suppied, use these areas as boundaries.  Otherwise, calculate
        boundaries using `segmentation.find_boundaries`

    See Also
    --------
    skimage.segmentation.find_boundaries

    """
    if connectivity is None:
        connectivity = np.asarray(masks[0]).ndim

    return [
        segmentation.find_boundaries(m, connectivity=connectivity, mode=mode, **kws)
        for m in masks
    ]


def find_boundaries_overlap(
    masks,
    boundaries=None,
    flag_none=True,
    mode="thick",
    connectivity=None,
    method="exact",
):
    """
    Find regions where boundaries overlap

    Parameters
    ----------
    masks : list of arrays of bool
        Convesions is `masks[i][index] == True`` if `index` is active for the ith mask
    boundaries : list of arrays of bool, optional
        If suppied, use these areas as boundaries.  Otherwise, calculate
        boundaries using `find_boundaries`
    flag_none : bool, default=True
        if True, replace overlap with None if no overlap between regions
    method : str, {'approx', 'exact'}
        * approx : consider regions where boundaries overlap and in one of the two region
        * exact : consider regions where boundaries overlap

    Returns
    -------
    overlap : dict of masks
        overlap[i, j] = region of overlap between boundaries and masks in regions i and j


    See Also
    --------
    find_boundaries
    """

    n = len(masks)
    assert method in ["approx", "exact"]

    if boundaries is None:
        boundaries = find_boundaries(masks, mode=mode, connectivity=connectivity)

    assert n == len(boundaries)

    result = {}
    for i, j in itertools.combinations(range(n), 2):
        if method == "approx":
            # overlap region is where boundaries overlap, and in one of the masks
            overlap = (boundaries[i] & boundaries[j]) & (masks[i] | masks[j])
            if flag_none and not np.any(overlap):
                overlap = None
            result[i, j] = overlap
        elif method == "exact":
            boundaries_overlap = boundaries[i] & boundaries[j]
            for index, m in enumerate([masks[i], masks[j]]):
                overlap = boundaries_overlap & m
                if flag_none and not np.any(overlap):
                    overlap = None
                result[i, j, index] = overlap

    return result


@docfiller_shared
def find_masked_extrema(
    data,
    masks,
    convention="image",
    extrema="max",
    fill_val=np.nan,
    fill_arg=None,
    unravel=True,
):
    """
    Find position and value of extrema of masked data.

    Parameters
    ----------
    data : ndarray
        input data to consider
    masks : list of bool arrays
        masks for data regions
    {mask_convention}
    extrema : {{'max','min}}
        Type of extrema to calculate
    fill_val : scalar, default=nan
        Value to fill for `out_arg` for empty `mask`
    fill_arg : scalar, optional
        Value to fill for `out_val` for empty `mask`
    unravel : bool, default=True
        If True, unravel flat index to multi dimensional index.


    Returns
    -------
    out_arg : list of indices
        index of extrema, one for each `mask`
    out_val : ndarray
        value of extrema, one for each `mask`
    """

    if extrema == "max":
        func = np.argmax
    elif extrema == "min":
        func = np.argmin
    else:
        raise ValueError('extrema must be on of {"min", "max}')

    masks = masks_change_convention(masks, convention, "image")

    data_flat = data.reshape(-1)
    positions_flat = np.arange(data.size)

    out_val = []
    out_arg = []

    for mask in masks:

        if mask is None or not np.any(mask):
            arg = fill_arg
            val = fill_val
        else:
            mask_flat = mask.reshape(-1)
            arg = positions_flat[mask_flat][func(data_flat[mask_flat])]
            val = data_flat[arg]

            if unravel:
                arg = np.unravel_index(arg, data.shape)

        out_arg.append(arg)
        out_val.append(val)

    out_val = np.array(out_val)

    return out_arg, out_val


@docfiller_shared
def merge_regions(
    w_tran,
    w_min,
    masks,
    nfeature_max=None,
    efac=1.0,
    force=True,
    convention="image",
    warn=True,
):
    r"""
    Merge labels where free energy energy barrier < efac.

    The scaled free energy :math:`w = \beta f = - \ln \Pi ` is analized.
    Anywhere where :math:`\Delta w = \text{{w_tran}} - \text{{w_min}} < \text{{efac}}` will be merged into a single phase.  Here,
    ``w_trans`` is the transition energy between phases (i.e., the minimum value of `lnPi`` between regions`) and ``w_min`` is the minimum energy (the maximum `lnPi`) for a phase.



    Parameters
    ----------
    w_tran : array
        Shape of array is ``(n, n)``, where ``n`` is the number of unique regions/phases.
    This is the transitional free energy ()
    w_min : array
        Shape of array is ``(n, 1)``.
    masks : sequence of bool arrays
        Masks for each region using `convention`
    nfeature_max : int
        maximum number of features/phases to allow
    efac : float, default=0.5
        Wnergy difference to merge on. When ``w_trans[i, j] - w_min[i] < efac``, phases
        ``i`` and ``j`` will be merged together.
    force : bool, default=True
        if True, then keep going until nfeature <= nfeature_max
        even if min_val > efac.
    {mask_convention}
    warn : bool, default=True
        if True, give warning messages

    Returns
    -------
    masks : list of bool arrays
        output masks using `convention`
    w_trans : array
        transition energy for new masks
    w_min : array
        free energy minima of new masks
    """

    nfeature = len(masks)
    if nfeature_max is None:
        nfeature_max = nfeature

    w_tran = np.array(w_tran, copy=True)
    w_min = np.array(w_min, copy=True)

    # keep track of keep/kill
    mapping = {i: msk for i, msk in enumerate(masks)}
    for cnt in range(nfeature):
        # number of finite minima
        nfeature = len(mapping)

        de = w_tran - w_min

        min_arg = np.unravel_index(np.nanargmin(de), de.shape)
        min_val = de[min_arg]

        if min_val > efac:
            if not force:
                if warn and nfeature > nfeature_max:
                    warnings.warn(
                        "min_val > efac, but still too many phases",
                        Warning,
                        stacklevel=2,
                    )
                break
            elif nfeature <= nfeature_max:
                break

        idx_keep, idx_kill = min_arg
        # keep the one with lower energy
        if w_min[idx_keep, 0] > w_min[idx_kill, 0]:
            idx_keep, idx_kill = idx_kill, idx_keep

        # transition from idx_keep to any other phase equals the minimum transition
        # from either idx_keep or idx_kill to that other phase
        new_tran = w_tran[[idx_keep, idx_kill], :].min(axis=0)
        new_tran[idx_keep] = np.inf
        w_tran[idx_keep, :] = w_tran[:, idx_keep] = new_tran

        # get rid of old one
        w_tran[idx_kill, :] = w_tran[:, idx_kill] = np.inf

        # new mask
        mapping[idx_keep] |= mapping[idx_kill]
        del mapping[idx_kill]

    # from mapping create some new stuff
    # new w/de
    idx_min = list(mapping.keys())
    w_min = w_min[idx_min]

    idx_tran = np.ix_(*(idx_min,) * 2)
    w_tran = w_tran[idx_tran]

    # get masks
    masks = [mapping[i] for i in idx_min]

    # optionally convert image
    masks = masks_change_convention(masks, True, convention)

    return masks, w_tran, w_min


@docfiller_shared
class wFreeEnergy(object):
    r"""
    Analysis of local free energy :math:`w = \beta f = - \ln \Pi`.

    Parameters
    ----------
    data : ndarray
        lnPi data
    masks : sequence of arrays of bool
        `masks[i]` is the mask indicating the `ith` phase.  See `convention`.
    {mask_convention}
    connectivity : int, optional
        connectivity parameter for boundary construction
    index : sequence of ints, optional
        Optional index to apply to phases.
        Not yet fully supported.

    Notes
    -----
    this class uses the 'image' convention for masks.  That is
    where `mask == True` indicates locations in the phase/region,
    and `mask == False` indicates locations not in the phase/region.
    This is opposite the convention used in :mod:`numpy.ma`.
    """

    def __init__(self, data, masks, convention="image", connectivity=None, index=None):
        self.data = np.asarray(data)

        # make sure masks in image convention
        self.masks = masks_change_convention(masks, convention, "image")

        if index is None:
            index = np.arange(self.nfeature)
        self.index = index

        if connectivity is None:
            connectivity = self.data.ndim
        self.connectivity = connectivity

    @property
    def nfeature(self):
        """number of features/regions/phases"""
        return len(self.masks)

    @classmethod
    @docfiller_shared
    def from_labels(
        cls,
        data,
        labels,
        connectivity=None,
        features=None,
        include_boundary=False,
        **kwargs
    ):
        """
        create wFreeEnergy from labels

        Parameters
        ----------
        data : ndarray
            lnPi data
        {labels}
        connectivity : int, optional
            connectivity parameters
        features : array-like, optional
            list of features to extract from labels.  Note that returned
            mask[i] corresponds to labels == feature[i].
        include_boundary : bool, default=False
            if True, include boundary regions in output mask
        **kwargs
            Extra arguments to :func:`lnpy.utils.labels_to_masks`

        Returns
        -------
        out : wFreeEnergy instance

        See Also
        --------
        lnPi.utils.labels_to_masks
        """
        masks, features = labels_to_masks(
            labels,
            features=features,
            convention="image",
            include_boundary=include_boundary,
            **kwargs
        )
        return cls(data=data, masks=masks, connectivity=connectivity)

    @gcached()
    def _data_max(self):
        """
        for lnPi data, find absolute argmax and max
        """
        return find_masked_extrema(self.data, self.masks)

    @gcached(prop=False)
    def _boundary_max(self, method="exact"):
        """
        find argmax along boundaries of regions.
        Corresponds to argmin(w)

        if method == 'exact', then find the boundary of each region
        and find max.  then find min of those maxes.
        """
        overlap = find_boundaries_overlap(
            self.masks,
            mode="thick",
            connectivity=self.connectivity,
            flag_none=True,
            method=method,
        )
        argmax, valmax = find_masked_extrema(
            self.data,
            overlap.values(),
            fill_val=np.nan,
            fill_arg=None,
        )

        # unpack output
        out_arg = {}
        out_max = np.full((self.nfeature,) * 2, dtype=float, fill_value=np.nan)
        if method == "approx":
            for (i, j), arg, val in zip(overlap.keys(), argmax, valmax):
                out_max[i, j] = out_max[j, i] = val
                out_arg[i, j] = arg

        elif method == "exact":
            # attach keys to argmax, valmax
            argmax = dict(zip(overlap.keys(), argmax))
            valmax = dict(zip(overlap.keys(), valmax))

            # first get unique keys:
            keys = [(i, j) for i, j, _ in overlap.keys()]
            keys = list(set(keys))

            for (i, j) in keys:
                vals = [valmax[i, j, index] for index in range(2)]
                # take min value of maxes
                if np.all(np.isnan(vals)):
                    out_arg[i, j] = None
                else:
                    idx_min = np.nanargmin(vals)
                    out_arg[i, j] = argmax[i, j, idx_min]
                    out_max[i, j] = out_max[j, i] = valmax[i, j, idx_min]
        return out_arg, out_max

    @property
    def w_min(self):
        """Minimum value of `w` (max `lnPi`) in each phase/region."""
        return -self._data_max[1][:, None]

    @property
    def w_argmin(self):
        """locations of the minimum of `w` in each phase/region"""
        return self._data_max[0]

    @property
    def w_tran(self):
        """Minimum value of `w` (max `lnPi`) in the boundary between phases.

        `w_tran[i, j]` is the transition energy between phases `i` ans `j`.
        """
        return np.nan_to_num(-self._boundary_max()[1], nan=np.inf)

    @property
    def w_argtran(self):
        """location of `w_tran`"""
        return self._boundary_max()[0]

    @gcached()
    def delta_w(self):
        """Transition energy ``delta_w[i, j] = w_tran[i, j] - w_min[i]``."""
        return self.w_tran - self.w_min

    def merge_regions(
        self,
        nfeature_max=None,
        efac=1.0,
        force=True,
        convention="image",
        warn=True,
    ):
        """
        Merge labels where free energy energy barrier < efac.

        Interface to :func:`merge_regions`

        Parameters
        ----------
        nfeature_max : int
            maximum number of features/phases to allow
        efac : float, default=0.5
            Wnergy difference to merge on. When ``w_trans[i, j] - w_min[i] < efac``, phases
            ``i`` and ``j`` will be merged together.
        force : bool, default=True
            if True, then keep going until nfeature <= nfeature_max
            even if min_val > efac.
        {mask_convention}
        warn : bool, default=True
            if True, give warning messages

        Returns
        -------
        masks : list of bool arrays
            output masks using `convention`
        w_trans : array
            transition energy for new masks
        w_min : array
            free energy minima of new masks
        """

        return merge_regions(
            w_tran=self.w_tran,
            w_min=self.w_min,
            masks=self.masks,
            nfeature_max=nfeature_max,
            efac=efac,
            force=force,
            convention=convention,
            warn=warn,
        )


def _get_w_data(index, w):
    w_min = pd.Series(w.w_min[:, 0], index=index, name="w_min")
    w_argmin = pd.Series(w.w_argmin, index=w_min.index, name="w_argmin")

    w_tran = (
        pd.DataFrame(
            w.w_tran,
            index=index,
            columns=index.get_level_values("phase").rename("phase_nebr"),
        )
        .stack()
        .rename("w_tran")
    )

    # get argtrans values for each index
    index_map = {idx: i for i, idx in enumerate(index.get_level_values("phase"))}
    v = w.w_argtran

    argtran = []
    for idxs in zip(
        *[w_tran.index.get_level_values(_) for _ in ["phase", "phase_nebr"]]
    ):
        i, j = [index_map[_] for _ in idxs]

        if (i, j) in v:
            val = v[i, j]
        elif (j, i) in v:
            val = v[j, i]
        else:
            val = None
        argtran.append(val)

    w_argtran = pd.Series(argtran, index=w_tran.index, name="w_argtran")

    return {
        "w_min": w_min,
        "w_tran": w_tran,
        "w_argmin": w_argmin,
        "w_argtran": w_argtran,
    }  # [index_map, w.w_argtran]}


class wFreeEnergyCollection(object):
    r"""
    Calculate the transition free energies for a :class:`lnpy.lnPiCollection`.

    :math:`w(N) = \beta f(N) = - \ln \Pi(N)`

    Parameters
    ----------
    parent : lnPiCollection

    Notes
    -----
    An instance of :class:`wFreeEnergyCollection` is normally created from the accessor :meth:`lnpy.lnPiCollection.wfe`
    """

    def __init__(self, parent):
        self._parent = parent
        self._use_joblib = getattr(self._parent, "_use_joblib", False)

    def _get_items_ws(self):
        indexes = []
        ws = []
        for meta, phases in self._parent.groupby_allbut("phase"):
            indexes.append(phases.index)
            masks = [x.mask for x in phases.values]
            ws.append(
                wFreeEnergy(data=phases.iloc[0].data, masks=masks, convention=False)
            )
        return indexes, ws

    @gcached()
    def _data(self):
        indexes, ws = self._get_items_ws()
        seq = get_tqdm(zip(indexes, ws), total=len(ws), desc="wFreeEnergyCollection")
        out = parallel_map_func_starargs(
            _get_w_data, items=seq, use_joblib=self._use_joblib, total=len(ws)
        )

        result = {key: pd.concat([x[key] for x in out]) for key in out[0].keys()}

        return result

    @property
    def w_min(self):
        """Minimum energy (maximum `lnPi`) for a given region/phase"""
        return self._data["w_min"]

    @property
    def w_tran(self):
        """Minimum energy (maximum `lnPi`) at boundary between phases"""
        return self._data["w_tran"]

    @property
    def w_argmin(self):
        """location of :attr:`w_min`"""
        return self._data["w_argmin"]

    @property
    def w_argtran(self):
        """location of :attr:`w_tran`"""
        return self._data["w_argtran"]

    @property
    def dw(self):
        """Series representation of `dw = w_tran - w_min`"""
        return (self.w_tran - self.w_min).rename("delta_w")

    @property
    def dwx(self):
        """xarray representation of :attr:`dw`"""
        return self.dw.to_xarray()

    def get_dwx(self, idx, idx_nebr=None):
        """
        helper function to get the change in energy from
        phase idx to idx_nebr.

        Parameters
        ----------
        idx : int
            phase index to consider transitions from
        idx_nebr : int or list, optional
            if supplied, consider transition from idx to idx_nebr or minimum of all element in idx_nebr.
            Default behavior is to return minimum transition from idx to all other neighboring regions

        Returns
        -------
        dw : DataArray
            Transition energy from `idx` to `idx_nebr`
            - if only phase idx exists, dw = np.inf
            - if idx does not exists, dw = 0.0 (no barrier between idx and anything else)
            - else min of transition for idx to idx_nebr
        """

        delta_w = self.dwx

        # reindex so that has idx in phase
        reindex = delta_w.indexes["phase"].union(pd.Index([idx], name="phase"))
        delta_w = delta_w.reindex(phase=reindex, phase_nebr=reindex)

        # much simpler
        if idx_nebr is None:
            delta_w = delta_w.sel(phase=idx)
        else:
            if not isinstance(idx_nebr, list):
                idx_nebr = [idx_nebr]
            if idx not in idx_nebr:
                idx_nebr.append(idx)
            nebrs = delta_w.indexes["phase_nebr"].intersection(idx_nebr)
            delta_w = delta_w.sel(phase=idx, phase_nebr=nebrs)

        out = delta_w.min("phase_nebr").fillna(0.0)
        return out

    def get_dw(self, idx, idx_nebr=None):
        """Sereis version of :meth:`get_dwx`"""
        return self.get_dwx(idx, idx_nebr).to_series()


# @lnPiCollection.decorate_accessor("wfe_phases")
class wFreeEnergyPhases(wFreeEnergyCollection):
    """
    Stripped down version of :class:`wFreeEnergyCollection` for single phase grouping.

    This should be used for a collection of lnPi that is at a single state point, with multiple phases.

    Parameters
    ----------
    parent : lnPiCollection

    Notes
    -----
    This is accessed through :attr:`lnpy.lnPiCollection.wfe_phases`

    """

    @gcached()
    def dwx(self):
        index = list(self._parent.index.get_level_values("phase"))
        masks = [x.mask for x in self._parent]
        w = wFreeEnergy(data=self._parent.iloc[0].data, masks=masks, convention=False)

        dw = w.w_tran - w.w_min
        dims = ["phase", "phase_nebr"]
        coords = dict(zip(dims, [index] * 2))
        return xr.DataArray(dw, dims=dims, coords=coords)

    @gcached()
    def dw(self):
        """Series representation of delta_w"""
        return self.dwx.to_series()

    def get_dw(self, idx, idx_nebr=None):
        dw = self.dwx
        index = dw.indexes["phase"]

        if idx not in index:
            return 0.0
        elif idx_nebr is None:
            nebrs = index.drop(idx)
        else:
            if not isinstance(idx_nebr, list):
                idx_nebr = [idx_nebr]
            nebrs = [x for x in idx_nebr if x in index]

        if len(nebrs) == 0:
            return np.inf
        return dw.sel(phase=idx, phase_nebr=nebrs).min("phase_nebr").values


@lnPiCollection.decorate_accessor("wfe")
def wfe_accessor(parent):
    """
    Accessor to :class:`~lnpy.wFreeEnergyCollection` from `self.wfe`.
    """
    return wFreeEnergyCollection(parent)


@lnPiCollection.decorate_accessor("wfe_phases")
def wfe_phases_accessor(parent):
    """Accessor to :class:`~lnpy.wFreeEnergyPhases` from `self.wfe_phases`."""
    return wFreeEnergyPhases(parent)


from warnings import warn


# create alias accessors
@lnPiCollection.decorate_accessor("wlnPi")
def wlnPi_accessor(parent):
    """Deprecated accessor to :class:`~lnpy.wFreeEnergyCollection` from `self.wlnPi`.

    Alias to `self.wfe`
    """
    warn("Using `wlnPi` accessor is deprecated.  Please use `wfe` accessor instead")
    return parent.wfe


@lnPiCollection.decorate_accessor("wlnPi_single")
def wlnPi_single_accessor(parent):
    """Deprecated accessor to :class:`~lnpy.wFreeEnergyPhases` from `self.wlnPi_single`.

    Alias to `self.wfe_single`
    """
    warn("Using `wlnPi_single is deprecated.  Please use `self.wfe_phases` instead")
    return parent.wfe_phases



"""
Set of routines to bracket solution of "thing"
"""


# def _bracket(
#     p_start,
#     build_phases,
#     check_condition,
#     sign=+1,
#     dlnz=0.5,
#     dfac=1.0,
#     ntry=20,
#     ref=None,
#     build_kws=None,
# ):

#     if build_kws is None:
#         build_kws = {}

#     new_lnz = p_start._get_lnz(build_phases.index)
#     dlnz_ = dlnz
#     for i in range(ntry):
#         new_lnz += sign * dlnz_
#         p = build_phases(new_lnz, ref=ref, **build_kws)
#         done, new_lnz, dlnz_ = check_condition(p, new_lnz, dlnz_)
#         if done:
#             return p, ntry

#         dlnz_ *= dfac

#     return None, ntry


# def _initial_bracket_spinodal_right(
#     C,
#     build_phases,
#     idx,
#     idx_nebr=None,
#     efac=1.0,
#     dlnz=0.5,
#     dfac=1.0,
#     vmax=1e5,
#     ntry=20,
#     step=+1,
#     ref=None,
#     build_kws=None,
# ):

#     if build_kws is None:
#         build_kws = {}
#     if efac <= 0:
#         raise ValueError("efac must be positive")

#     # Series representation of dw
#     s = C.wfe.get_dw(idx, idx_nebr)
#     if step < 0:
#         s = s.iloc[-1::-1]

#     # left bounds
#     s_idx = s[s > 0.0]
#     if len(s_idx) == 0:
#         raise ValueError("no phase {}".format(idx))
#     ss = s_idx[s_idx > efac]
#     if len(ss) > 0:
#         left = C.mloc[ss.index[[-1]]]
#     else:

#         def condition_left(p, new_lnz, dlnz_):
#             done = (
#                 idx in p._get_leve("phase")
#                 and p.wfe_signle.get_dw(idx, idx_nebr) > efac
#             )
#             return done, new_lnz, dlnz_

#         left, left_ntry = _bracket(
#             C.mloc[s_idx.index[[0]]],
#             build_phases,
#             condition_left,
#             sign=-1 * step,
#             dlnz=dlnz,
#             dfac=dfac,
#             ntry=ntry,
#             ref=ref,
#             build_kws=build_kws,
#         )
#     if left is None:
#         raise RuntimeError("could not find left")

#     # right
#     ss = s[s < efac]
#     if len(ss) > 0:
#         right = C.mloc[ss.index[[0]]]
#     else:

#         def condition_right(p, new_llnz, dlnz_):
#             done = (
#                 idx not in p._get_leve("phase")
#                 or t.wfe_phases.get_dw(idx, idx_nebr) < efac
#             )
#             return new, new_lnz, dlnz_

#         right, right_ntry = _bracket(
#             C.mloc[s.index[[-1]]],
#             build_phases,
#             condition_right,
#             sign=+1 * step,
#             dlnz=dlnz,
#             dfac=dfac,
#             ntry=ntry,
#             ref=ref,
#             build_kws=build_kws,
#         )



"""
Routines to load/run examples
"""

try:
    import importlib_resources as resources
except ImportError:
    import importlib.resources as resources

import json
from collections.abc import Callable
from dataclasses import asdict, dataclass

import numpy as np
import xarray as xr

# from .lnpiseries import lnPiCollection
from .lnpidata import lnPiMasked
from .segment import PhaseCreator
from .utils import dataset_to_lnpimasked


def json_to_dict(basename):
    """Load a json file into a dict.

    All files names are relative to 'lnpy/data/'

    Parameters
    ----------
    basename : string
        Name of file to load in 'lnpy/data' directory.

    Returns
    -------
    output : dict

    """

    if basename.endswith(".gz"):
        import gzip

        fopen = gzip.open
    else:
        fopen = open

    with fopen(resources.files("lnpy.data").joinpath(basename), "r") as f:
        out = json.load(f)

    return out


def load_example_dict(name):
    """
    Load a dictionary of data

    Parameters
    ----------
    name : {'lj_sub', 'lj_sup', 'ljmix_sup', 'hsmix', 'watermof}
    """

    ref = load_example_maskddata(name)

    return {
        "lnPi_data": ref.data,
        "lnPi_mask": ref.mask,
        "state_kws": ref.state_kws,
        "extra_kws": ref.extra_kws,
        "lnz": ref.lnz,
    }


def load_example_maskddata(name):
    """
    Load an example file

    Parameters
    ----------
    name : {'lj_sub', 'lj_sup', 'ljmix_sup', 'hsmix', 'watermof}
    """

    extensions = {
        "lj_sub": "json",
        "lj_sup": "json",
        "ljmix_sup": "json.gz",
        "hsmix": "json.gz",
        "watermof": "json",
    }

    assert name in extensions

    basename = f"{name}_example.{extensions[name]}"

    d = json_to_dict(basename)
    ds = xr.Dataset.from_dict(d)

    ref = dataset_to_lnpimasked(ds)
    return ref


@dataclass
class Example:
    ref: lnPiMasked
    phase_creator: PhaseCreator
    build_phases: Callable

    def to_dict(self):
        return asdict(self)

    def unpack(self, keys=None):
        if keys is None:
            keys = ["ref", "phase_creator", "build_phases"]
        return (getattr(self, k) for k in keys)


def lj_sup_example():
    ref = load_example_maskddata("lj_sup")

    phase_creator = PhaseCreator(
        nmax=1,
        nmax_peak=1,
        ref=ref,
        merge_kws={"efac": 0.8},
        tag_phases=None,
    )

    build_phases = phase_creator.build_phases_mu([None])

    return Example(ref=ref, phase_creator=phase_creator, build_phases=build_phases)


def tag_phases_single_comp_simple(x):
    if len(x) > 2:
        raise ValueError("bad tag function")
    argmax0 = np.array([xx.local_argmax()[0] for xx in x])
    return np.where(argmax0 <= x[0].shape[0] / 2, 0, 1)


def lj_sub_example():

    ref = load_example_maskddata("lj_sub")

    phase_creator = PhaseCreator(
        nmax=2,
        nmax_peak=4,
        ref=ref,
        merge_kws={"efac": 0.8},
        tag_phases=tag_phases_single_comp_simple,
    )

    build_phases = phase_creator.build_phases_mu([None])

    return Example(ref=ref, phase_creator=phase_creator, build_phases=build_phases)


def ljmix_sup_example():
    ref = load_example_maskddata("ljmix_sup")

    phase_creator = PhaseCreator(
        nmax=1,
        ref=ref,
    )

    build_phases = phase_creator.build_phases
    return Example(ref=ref, phase_creator=phase_creator, build_phases=build_phases)


def hsmix_example():
    def tag_phases(x):
        if len(x) > 2:
            raise ValueError("bad tag function")
        argmax0 = np.array([xx.local_argmax()[0] for xx in x])
        return np.where(argmax0 <= x[0].shape[0] / 2, 0, 1)

    ref = load_example_maskddata("hsmix")

    phase_creator = PhaseCreator(
        nmax=2, nmax_peak=4, ref=ref, tag_phases=tag_phases, merge_kws={"efac": 0.8}
    )

    return Example(ref=ref, phase_creator=phase_creator, build_phases=None)


# def watermof_example():
#     def tag_phases(x):
#         if len(x) > 2:
#             raise ValueError("bad tag function")
#         argmax0 = np.array([xx.local_argmax()[0] for xx in x])
#         return np.where(argmax0 <= x[0].shape[0] / 2, 0, 1)

#     ref = load_example_maskddata('watermof')

#     phase_creator = PhaseCreator(
#         nmax=2,
#         nmax_peak=10,
#         merge_kws={'efac': 0.8}

#     )



"""
define optional paremeters
"""

import os

TQDM_USE = "tqdm_use"
TQDM_LEN_CALC = "tqdm_len_calc"
TQDM_LEN_BUILD = "tqdm_len_build"
TQDM_LEAVE = "tqdm_leave"
TQDM_BAR = "tqdm_bar"

JOBLIB_USE = "joblib_use"
JOBLIB_N_JOBS = "joblib_n_jobs"
JOBLIB_BACKEND = "joblib_backend"
JOBLIB_KWS = "joblib_kws"
JOBLIB_LEN_CALC = "joblib_len_calc"
JOBLIB_LEN_BUILD = "joblib_len_build"

DOC_SUB = "doc_sub"

try:
    # Default to setting docs
    _DOC_SUB = os.getenv("LNPI_DOC_SUB", "True").lower() not in ("0", "f", "false")
except KeyError:
    _DOC_SUB = True


OPTIONS = {
    TQDM_USE: True,
    TQDM_LEN_CALC: 100,
    TQDM_LEN_BUILD: 500,
    TQDM_LEAVE: False,
    TQDM_BAR: "default",
    JOBLIB_USE: True,
    JOBLIB_N_JOBS: -1,
    JOBLIB_BACKEND: None,
    JOBLIB_KWS: {},
    JOBLIB_LEN_CALC: 200,
    JOBLIB_LEN_BUILD: 500,
    DOC_SUB: _DOC_SUB,
}

_isbool = lambda x: isinstance(x, bool)
_isint = lambda x: isinstance(x, int)
_isstr = lambda x: isinstance(x, str)

_isstr_or_None = lambda x: isinstance(x, str) or x is None

_VALIDATORS = {
    TQDM_USE: _isbool,
    TQDM_LEN_CALC: _isint,
    TQDM_LEN_BUILD: _isint,
    TQDM_LEAVE: _isbool,
    TQDM_BAR: lambda x: x in ["default", "text", "notebook"],
    JOBLIB_USE: _isbool,
    JOBLIB_N_JOBS: _isint,
    JOBLIB_BACKEND: _isstr_or_None,
    JOBLIB_KWS: lambda x: isinstance(x, dict),
    JOBLIB_LEN_CALC: _isint,
    JOBLIB_LEN_BUILD: _isint,
    DOC_SUB: _isbool,
}

_SETTERS = {}


class set_options(object):
    """Set options for xarray in a controlled context.
    Currently supported options:
    - `tqdm_use` : if `True`, use progress bar where appropriate
    - `tqdm_len_calc` : min length for using bar in calculations of properties
    - `tqdm_len_build` : min length for building Collection objects
    - `tqdm_leave` : if True, leave bar after execution.  Default=False
    - `joblib_use` : if `True`, use joblib
    - `joblib_n_jobs` : number of processors to use, default=all processors
    - `joblib_backend` : backend to use.  Default='multiprocessing'.
    - `joblib_kws` : dictionary of arguments to joblib.Parallel.
    - `joblib_len_calc` : min length to use parallel in calculations
    - `joblib_len_build` : min lenght to use parallel in build
    You can use ``set_options`` either as a context manager:
    >>> with xr.set_options(use_tqdm=True, tqdm_min_len_calc=50):
    ...     c.xge.betaOmega()
    ...
    Or to set global options:
    >>> xr.set_options(tqdm_min_len_calc=50)
    """

    def __init__(self, **kwargs):
        self.old = {}
        for k, v in kwargs.items():
            if k not in OPTIONS:
                raise ValueError(
                    "argument name %r is not in the set of valid options %r"
                    % (k, set(OPTIONS))
                )
            if k in _VALIDATORS and not _VALIDATORS[k](v):
                raise ValueError(f"option {k!r} given an invalid value: {v!r}")
            self.old[k] = OPTIONS[k]
        self._apply_update(kwargs)

    def _apply_update(self, options_dict):
        for k, v in options_dict.items():
            if k in _SETTERS:
                _SETTERS[k](v)
        OPTIONS.update(options_dict)

    def __enter__(self):
        return

    def __exit__(self, type, value, traceback):
        self._apply_update(self.old)
